from ..local.generic import Generic

test01 = Generic(
    "TEST01"
    ,{"cutoff":"2021-03-01","closing":"2021-04-01","firstPay":"2021-06-20"
     ,"payFreq":["DayOfMonth",20],"poolFreq":"MonthFirst","stated":"2030-01-01"}
    ,{'assets':[["Mortgage"
        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":20
          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
          ,{"currentBalance":2200
          ,"currentRate":0.08
          ,"remainTerm":20
          ,"status":"current"}]]}
    ,(("acc01",{"balance":0}),)
    ,(("A1",{"balance":1000
             ,"rate":0.07
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.08}
             ,"bondType":{"Sequential":None}})
      ,("B",{"balance":1000
             ,"rate":0.0
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.00}
             ,"bondType":{"Equity":None}
             }))
    ,tuple()
    ,{"amortizing":[
         ["accrueAndPayInt","acc01",["A1"]]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payIntResidual","acc01","B"]
     ]}
    ,[["CollectedInterest","acc01"]
      ,["CollectedPrincipal","acc01"]
      ,["CollectedPrepayment","acc01"]
      ,["CollectedRecoveries","acc01"]]
    ,None
    ,None
    ,None
    ,None
    ,("PreClosing","Amortizing")
    )


test02 = Generic(
    "TEST01"
    ,{"cutoff":"2021-03-01","closing":"2021-04-01","firstPay":"2021-06-20"
     ,"payFreq":["DayOfMonth",20],"poolFreq":"MonthFirst","stated":"2030-01-01"}
    ,{'assets':[["Mortgage"
        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":20
          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
          ,{"currentBalance":2200
          ,"currentRate":0.08
          ,"remainTerm":20
          ,"status":"current"}]]}
    ,(("acc01",{"balance":0}),)
    ,(("A1",{"balance":800
             ,"rate":0.07
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.08}
             ,"bondType":{"Sequential":None}})
      ,("B",{"balance":1000
             ,"rate":0.0
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.00}
             ,"bondType":{"Equity":None}
             }))
    ,tuple()
    ,{"amortizing":[
         ["accrueAndPayInt","acc01",["A1"]]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payIntResidual","acc01","B"]
     ]}
    ,[["CollectedInterest","acc01"]
      ,["CollectedPrincipal","acc01"]
      ,["CollectedPrepayment","acc01"]
      ,["CollectedRecoveries","acc01"]]
    ,None
    ,None
    ,None
    ,None
    ,("PreClosing","Amortizing")
    )

test03 =  Generic(
    "TEST03:two fees"
    ,{"cutoff":"2021-03-01","closing":"2021-04-01","firstPay":"2021-06-20"
     ,"payFreq":["DayOfMonth",20],"poolFreq":"MonthFirst","stated":"2030-01-01"}
    ,{'assets':[["Mortgage"
        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":20
          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
          ,{"currentBalance":2200
          ,"currentRate":0.08
          ,"remainTerm":20
          ,"status":"current"}]]}
    ,(("acc01",{"balance":0}),)
    ,(("A1",{"balance":800
             ,"rate":0.07
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.08}
             ,"bondType":{"Sequential":None}})
      ,("B",{"balance":1000
             ,"rate":0.0
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.00}
             ,"bondType":{"Equity":None}
             }))
    ,(("bondAdmFee",{"type":{"pctFee":[("bondBalance",),0.02]},"feeStart":"2021-04-01"})
      ,("poolServFee",{"type":{"annualPctFee":[("poolBalance",),0.02]},"feeStart":"2021-04-01"}))
    ,{"amortizing":[
         ["calcAndPayFee","acc01",["bondAdmFee","poolServFee"]]
         ,["accrueAndPayInt","acc01",["A1"]]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payIntResidual","acc01","B"]
     ]}
    ,[["CollectedInterest","acc01"]
      ,["CollectedPrincipal","acc01"]
      ,["CollectedPrepayment","acc01"]
      ,["CollectedRecoveries","acc01"]]
    ,None
    ,None
    ,None
    ,None
    ,("PreClosing","Amortizing")
    )


test04 =  Generic(
    "TEST04: ongoing deal with defaults"
    ,{"pay":["2021-09-20","2021-10-20"]
      ,"collect":["2021-09-30","2021-10-31"]
     ,"payFreq":["DayOfMonth",20],"poolFreq":"MonthFirst","stated":"2030-01-01"}
    ,{'assets':[["Mortgage"
        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":20
          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
          ,{"currentBalance":2200
          ,"currentRate":0.08
          ,"remainTerm":20
          ,"status":"current"}]]
       ,'issuanceStat':{"HistoryDefaults":150}}
    ,(("acc01",{"balance":0}),)
    ,(("A1",{"balance":800
             ,"rate":0.07
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.08}
             ,"bondType":{"Sequential":None}})
      ,("B",{"balance":1000
             ,"rate":0.0
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.00}
             ,"bondType":{"Equity":None}
             }))
    ,(("bondAdmFee",{"type":{"pctFee":[("bondBalance",),0.02]},"feeStart":"2021-09-20"})
      ,("poolServFee",{"type":{"annualPctFee":[("poolBalance",),0.02]},"feeStart":"2021-09-20"}))
    ,{"amortizing":[
         ["calcAndPayFee","acc01",["bondAdmFee","poolServFee"]]
         ,["accrueAndPayInt","acc01",["A1"]]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payIntResidual","acc01","B"]
     ]}
    ,[["CollectedInterest","acc01"]
      ,["CollectedPrincipal","acc01"]
      ,["CollectedPrepayment","acc01"]
      ,["CollectedRecoveries","acc01"]]
    ,None
    ,None
    ,None
    ,None
    ,"Amortizing"
    )

test05 = Generic(
    "debug example"
    ,{"cutoff":"2021-03-01","closing":"2021-04-01","firstPay":"2021-06-20"
     ,"payFreq":["DayOfMonth",20],"poolFreq":"MonthFirst","stated":"2030-01-01"}
    ,{'assets':[["Mortgage"
        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":20
          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
          ,{"currentBalance":2200
          ,"currentRate":0.08
          ,"remainTerm":20
          ,"status":"current"}]]}
    ,(("acc01",{"balance":0}),)
    ,(("A1",{"balance":1000
             ,"rate":0.07
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.08}
             ,"bondType":{"Sequential":None}})
      ,("B",{"balance":1000
             ,"rate":0.0
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.00}
             ,"bondType":{"Equity":None}
             }))
    ,tuple()
    ,{"amortizing":[
         ["calcInt","A1"]
         ,["inspect","After AccureInteret bond:A1",("bondDueInt","A1")]
         ,["payInt","acc01",["A1"]]
         ,["inspect","After Pay",("bondDueInt","A1")]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payIntResidual","acc01","B"]
     ]}
    ,[["CollectedInterest","acc01"]
      ,["CollectedPrincipal","acc01"]
      ,["CollectedPrepayment","acc01"]
      ,["CollectedRecoveries","acc01"]]
    ,None
    ,None
    ,None
    ,None
    ,("PreClosing","Amortizing")
    )



import toolz as tz
from lenses import lens
import pandas as pd
import json, enum, os, pathlib, re
from htpy import body, h1, head, html, li, title, ul, div, span, h3, h2, a, h4,h5, h6
from markupsafe import Markup
from .local.cf import readInspect
from .local.cf import readBondsCf,readFeesCf,readAccsCf,readPoolsCf,readLedgers,readTriggers

class OutputType(int, enum.Enum):
    """Internal
    """
    Plain = 0 # print cashflow to html tables 
    Anchor = 1 # print cashflow with book mark
    Tabbed = 2 # TBD


def mapToList(m:dict, title_=h3, anchor=False):
    m2 = {}
    if not anchor:
        m2 = {title_[k]:div[Markup(v.to_html())] for k,v in m.items() }
    else:
        m2 = {title_(id=f"anchor-{anchor}-{k}")[k]:div[Markup(v.to_html())]
                for k,v in m.items() }

    return list(m2.items())

def buildSection(lst:list, title_=(h2,h3), anchor=False):
    return [ div[ title_[0](id=f"anchor-{_t}")[_t], mapToList(x, anchor=_t)] 
            for (_t, x) in lst]

def buildSectionFlat(lst:list, title_=h2, anchor=False):
    return [ div[ title_(id=f"anchor-{_t}")[_t],Markup(x.to_html())] 
            for (_t, x) in lst if x is not None]


def consolResp(r:dict):
    bondDf = ("Bond", tz.valfilter(lambda x: isinstance(x, pd.DataFrame), r['bonds']))
    bondGrpDf = ("BondGroup", 
                  tz.pipe(tz.valfilter(lambda x: not isinstance(x, pd.DataFrame), r['bonds'])
                          ,lambda x : {f"{k}-{k2}":v2 for k,v in x.items()
                                         for k2,v2 in v.items()}))
    poolDf = ("Pool", r['pool']['flow'])
    accDf = ("Accounts", r['accounts'])
    feeDf = ("Fee", r['fees'])

    reportDf = ("Finanical Reports", r['result'].get('report', {}))
    
    results = dict([("Status",r['result']['status'])
                    ,("Pricing",r["pricing"])
                    ,("Bond Summary",r['result']['bonds'])
                    ,("Log",r['result']['logs'])
                    ,("Waterfall",r['result']['waterfall'])
                    ,("Inspect",readInspect(r['result']))])
    

    jointDfs =[("MultiFee",readFeesCf(feeDf[1]))
               ,("MultiBond",readBondsCf(bondDf[1],popColumns=[]))
               ,("MultiAccounts",readAccsCf(accDf[1]))
               ,("MultiPools", readPoolsCf(poolDf[1]))
               ,("MultiLedger", readLedgers(r.get("ledgers",{})))
               ,("MultiTrigger", readTriggers(r.get("triggers",{})))]
    return dict(
        [bondDf]+[bondGrpDf]+[poolDf]+[accDf]+[feeDf]\
        +[reportDf]+[("result",results)]+[ (x[0],{x[0]:x[1]}) for x in jointDfs ]
    )

def toHtml(r:dict, p:str, style=OutputType.Plain, debug=False):
    """
    r : must be a result from "read=True"
    """

    bondDf = ("Bond", tz.valfilter(lambda x: isinstance(x, pd.DataFrame), r['bonds']))
    bondGrpDf = ("BondGroup", 
                 tz.pipe(tz.valfilter(lambda x: not isinstance(x, pd.DataFrame), r['bonds'])
                         ,lambda x : {f"{k}-{k2}":v2 for k,v in x.items()
                                        for k2,v2 in v.items()}
                         )
                )

    section0 = buildSection([bondDf,bondGrpDf])

    # section 1
    poolDf = ("Pool", r['pool']['flow'])
    accDf = ("Accounts", r['accounts'])
    feeDf = ("Fee", r['fees'])
    section1 = buildSection([poolDf, feeDf, accDf])

    # section 2
    section2 = buildSectionFlat([("Status",r['result']['status'])
                               ,("Pricing",r["pricing"])
                               ,("Bond Summary",r['result']['bonds'])
                               ,("Log",r['result']['logs'])
                               ,("Waterfall",r['result']['waterfall'])
                               ,("Inspect",readInspect(r['result']))])
    # section 3
    #section3 = [div[h2(id=f"anchor-{_t}")[_t], mapToList(x)]
    #            for (_t, x) in [("Finanical Reports", r['result'].get('report', {}))]]
    section3 = buildSection([("Finanical Reports", r['result'].get('report', {}))])

    # read joint cashflows 
    # section 4
    ledgerDf = ("Ledger", r.get("ledgers",{}))
    section4 = buildSectionFlat([("MultiFee",readFeesCf(feeDf[1]))
                               ,("MultiBond",readBondsCf(bondDf[1],popColumns=[]))
                               ,("MultiAccounts",readAccsCf(accDf[1]))
                               ,("MultiPools", readPoolsCf(poolDf[1]))
                               ,("MultiLedger", readLedgers(ledgerDf[1]))
                               ,("MultiTrigger", readTriggers(r.get("triggers",{})))
                               ])


    dealName = r['_deal']['contents']['name']
    c = html[
        head[title[dealName]],
        body[section0, section1, section2, section3 ,section4],
    ]

    if debug:
        return c

    if style == OutputType.Anchor:
        links = c & lens._children[1]._children.Each().Each()._children[0]._attrs.Regex(r'(?<=id=").*(?=")').collect()
        c &= lens._children[1]._children.Each().Each()._children[0]._children.modify(lambda x:[x,a(href="#toc")["  ^Top"] ])

        linksStr = [ l.lstrip("anchor-")  for l in links ]
        hrefs = ul[ [li[a(href=f"#{l}")[f" < {lstr} > "]] for (lstr,l) in zip(linksStr,links) ]]


        c &= lens._children[1]._children.modify(lambda xs: tuple([div(id="toc")[div["Table Of Content"],hrefs]])+xs)

    absPath = None
    with open(p, 'wb') as f:
        f.write(c.encode())
        absPath = os.path.abspath(os.path.join(os.getcwd(),p))

    return absPath


def toExcel(r:dict, p:str,exlude=[],headerFormat={'bold': True, 'bg_color': '#9fdd92','align':'center'}):
    x = consolResp(r)

    #filter out empty in m

    def annotateLoc(m:dict, skip=3):
        assert isinstance(m, dict),"annotateLoc must be a dict"
        if not m:
            return {}
        _r = {}
        beginRow = skip
        for k,v in m.items():
            if (v is None) or v.empty:
                continue
            _r[k] = beginRow
            beginRow += v.index.size + skip
        return _r

    locMap = tz.valmap(annotateLoc, x)

    with pd.ExcelWriter(p, mode='w', engine='xlsxwriter') as writer: 
        workbook = writer.book
        hFormat = workbook.add_format(headerFormat)
        for cat, m in x.items():
            if cat in exlude:
                continue
            workbook.add_worksheet(cat)
            currentSheet = writer.sheets[cat]
            for k,v in m.items():
                if v is None:
                    continue
                if v.empty:
                    continue
                currentSheet.write(locMap[cat][k]-1, 0, k, hFormat)
                v.to_excel(writer, sheet_name=cat, startrow=locMap[cat][k])
            currentSheet.autofit()

    return os.path.abspath(os.path.join(os.getcwd(),p))






import enum 
from .util import *

china_date = "日期"
english_date = "Date"


# Bond 
china_bondflow_cash = "本息合计"
china_bondflow_fields_s = ["余额", "利息", "本金", "执行利率", china_bondflow_cash,"应付利息","罚息" "本金系数", "备注"]
china_bondflow_fields = [china_date] + china_bondflow_fields_s
china_bond_cashflow = ["本金", "利息", china_bondflow_cash]

english_bondflow_cash = "cash"
english_bondflow_fields_s = ["balance", "interest", "principal", "rate", english_bondflow_cash, "intDue","intOverInt","factor",  "memo"]
english_bondflow_fields = ["date"] + english_bondflow_fields_s
english_bond_cashflow = ["principal", "interest", english_bondflow_cash]

# Cumulative
china_cumStats = ["累计还款", "累计早偿", "累计拖欠", "累计违约", "累计回收", "累计损失"]
english_cumStats = ["CumPrincipal", "CumPrepay", "CumDelinq", "CumDefault", "CumRecovery", "CumLoss"]

# Pool

## Non Balance flow
china_non_balance_flow = [ "利率","债务人数量","早偿手续费"]
english_non_balance_flow = [ "WAC","BorrowerNum","PrepayPenalty"]

## mortgage
china_mortgage_flow_fields = ["余额", "本金", "利息", "早偿金额", "违约金额", "回收金额", "损失金额"]+ china_non_balance_flow
english_mortgage_flow_fields = ["Balance", "Principal", "Interest", "Prepayment", "Default", "Recovery", "Loss"]+english_non_balance_flow

china_mortgage_delinq_flow_fields = ["余额", "本金", "利息", "早偿金额", "拖欠金额", "违约金额", "回收金额", "损失金额"]+ china_non_balance_flow
english_mortgage_delinq_flow_fields = ["Balance", "Principal", "Interest", "Prepayment", "Delinquency", "Default", "Recovery", "Loss"]+english_non_balance_flow

china_mortgage_flow_fields_d = [china_date]+ china_mortgage_flow_fields
english_mortgage_flow_fields_d = [english_date] + english_mortgage_flow_fields

china_mortgage_delinq_flow_fields_d = [china_date] + china_mortgage_delinq_flow_fields
english_mortgage_delinq_flow_fields_d = [english_date] + english_mortgage_delinq_flow_fields

## Rental/Lease flow
china_rental_flow = ['待收金额', '租金', '违约']
english_rental_flow = ['Balance', 'Rental', 'Default']
china_rental_flow_d = [china_date] + china_rental_flow
english_rental_flow_d = [english_date] + english_rental_flow

## Loan flow
china_loan_flow = ["余额", "本金", "利息", "早偿金额", "违约金额", "回收金额", "损失金额", "利率"]
english_loan_flow = ["Balance", "Principal", "Interest", "Prepayment", "Default", "Recovery", "Loss", "WAC"]
china_loan_flow_d = [china_date] + china_loan_flow
english_loan_flow_d = [english_date] + english_loan_flow

## Fixed flow 
china_fixed_flow = ["余额", "折旧", "累计折旧", "单元", "现金"]
english_fixed_flow = ["Balance", "Depreciation", "CumuDepreciation", "Unit", "Cash"]
china_fixed_flow_d = [china_date] + china_fixed_flow
english_fixed_flow_d = [english_date] + english_fixed_flow

## Receivable 
china_receivable_flow = ["余额", "应计费用", "本金", "费用", "违约", "回收", "损失"]
english_receivable_flow = ["Balance", "AccuredFee", "Principal", "Fee", "Default", "Recovery", "Loss"]
china_receivable_flow_d = [china_date] + china_receivable_flow
english_receivable_flow_d = [english_date] + english_receivable_flow

## Underlying Bond Flow
china_uBond_flow = ["余额", "本金", "利息"]
english_uBond_flow = ["Balance", "Principal", "Interest"]
china_uBond_flow_d = [china_date] + china_uBond_flow
english_uBond_flow_d = [english_date] + english_uBond_flow

# Fee 
china_fee_flow_fields_d = [china_date, "余额", "支付", "剩余支付", "备注"]
english_fee_flow_fields_d = ["date", "balance", "payment", "due", "memo"]

# Trigger 
china_trigger_flow_fields_d = [china_date, "status","备注"]
english_trigger_flow_fields_d = ["date","status","memo"]

# Account
china_acc_flow_fields_d = [china_date, "余额", "变动额", "备注"]
english_acc_flow_fields_d = ["date", "balance", "change", "memo"]


# LiqProvider
china_liq_flow_fields_d = [china_date, "额度" , "已提供", "利息", "费用", "现金", "备注"]
english_liq_flow_fields_d = ["date", "credit" , "used", "int", "premium", "cash", "memo"]

# Rate Swap 
china_rs_flow_fields_d = [china_date, "面额", "变动额", "支付", "收取", "净额", "备注"]
english_rs_flow_fields_d = ["date", "balance", "amount", "pay", "receive", "due", "memo"]

# Ledger
china_ledger_flow_fields_d = [china_date, "余额", "变动额", "备注"]
english_ledger_flow_fields_d = ["date", "balance", "amount", "comment"]

# deal status



datePattern = {"月末": "MonthEnd", "季度末": "QuarterEnd", "年末": "YearEnd", "月初": "MonthFirst",
               "季度初": "QuarterFirst", "年初": "YearFirst", "每年": "MonthDayOfYear", "每月": "DayOfMonth", "每周": "DayOfWeek"}

freqMap = {"每月": "Monthly", "每季度": "Quarterly", "每半年": "SemiAnnually", "每年": "Annually", "Monthly": "Monthly", "Quarterly": "Quarterly", "SemiAnnually": "SemiAnnually", "Annually": "Annually", "monthly": "Monthly", "quarterly": "Quarterly", "semiAnnually": "SemiAnnually", "annually": "Annually" 
           ,"Weekly":"Weekly","weekly":"Weekly","每周":"Weekly","BiWeekly":"BiWeekly","biweekly":"BiWeekly","每两周":"BiWeekly"}

rateLikeFormula = set(["bondFactor", "bondFactorOf", "poolFactor", "cumPoolDefaultedRate","poolWaRate","bondWaRate", "资产池累计违约率", "债券系数", "资产池系数"
                      ,"cumPoolNetLossRate","cumPoolDefaultedRateTill","比例","ratio","avgRatio"
                      ,"平均比例"])
intLikeFormula = set(["borrowerNumber", "monthsTillMaturity", "periodNum"])
boolLikeFormula = set(["trigger", "事件", "isMostSenior", "最优先", "isPaidOff","清偿完毕","rateTest","allTest","anyTest","比率测试","任一测试","所有测试"
                        "isOutstanding"])

bookDirection = {"Credit":"Credit" ,"Debit":"Debit"
                ,"借方":"Debit" ,"贷方":"Credit"
                ,"D":"Debit" ,"C":"Credit"
                ,"DR":"Debit" ,"CR":"Credit"
                ,"Dr":"Debit" ,"Cr":"Credit"}


op_map = {">": "G", ">=": "GE", "<": "L", "<=": "LE", "=": "E"}

#Deal Cycle
chinaDealCycle = {"回收后":"EndCollection"
                 ,"回收动作后":"EndCollectionWF"
                 ,"分配前":"BeginDistributionWF"
                 ,"分配后":"EndDistributionWF"
                 ,"分配中":"InWF"}

englishDealCycle = {"BeforeCollect":"EndCollection"
                   ,"AfterCollect":"EndCollectionWF"
                   ,"BeforeDistribution":"BeginDistributionWF"
                   ,"AfterDistribution":"EndDistributionWF"
                   ,"InDistribution":"InWF"}

dealCycleMap = chinaDealCycle | englishDealCycle

#Asset pricing 
assetPricingHeader = {
    "chinese":["估值", "WAL", "Duration", "Convexity", "AccruedInterest"]
    ,"english":["Pricing", "WAL", "Duration", "Convexity", "AccruedInterest"]
}

#account fields
accountHeader = {
    "chinese": {"idx": china_date, "change": "变动额", "bal": ("期初余额", '余额', "期末余额")}
    ,"english": {"idx": "date", "change": "change", "bal": ("begin balance", 'balance', "end balance")}
}


validCutoffFields = {
    "资产池规模": "IssuanceBalance"
    ,"发行规模": "IssuanceBalance"
    ,"CumulativeDefaults":"HistoryDefaults"
    ,"累计违约余额":"HistoryDefaults"
}

dealStatusLog = {'cn': [china_date, "旧状态", "新状态", "备注"]
                 ,'en': [english_date, "From", "To", "Comment"]}

dealStatusMap = {"en": {'amort': "Amortizing", 'def': "Defaulted", 'acc': "Accelerated", 'end': "Ended",
                        'called': "Called",
                        'pre': "PreClosing",'revol':"Revolving"
                        ,'warehousing':"Warehousing"}
                , "cn": {'amort': "摊销", 'def': "违约", 'acc': "加速清偿", 'end': "结束", 'pre': "设计","revol":"循环"
                        ,'called':"清仓回购"
                        ,'warehousing':"Warehousing"}}

cfIndexMap = {'cn':"日期",'en':english_date,"english":english_date,"chinese":"日期"}


class DC(enum.Enum):  # TODO need to check with HS code
    DC_30E_360 = "DC_30E_360"
    DC_30Ep_360 = "DC_30Ep_360"
    DC_ACT_360 = "DC_ACT_360"
    DC_ACT_365A = "DC_ACT_365A"
    DC_ACT_365L = "DC_ACT_365L"
    DC_NL_365 = "DC_NL_365"
    DC_ACT_365F = "DC_ACT_365F"
    DC_ACT_ACT = "DC_ACT_ACT"
    DC_30_360_ISDA = "DC_30_360_ISDA"
    DC_30_360_German = "DC_30_360_German"
    DC_30_360_US = "DC_30_360_US"


class InspectTags(str, enum.Enum):
    """ Inspect Tag when reading deal run logs """
    InspectBal = "InspectBal"
    InspectBool = "InspectBool"
    InspectRate = "InspectRate"
    InspectInt = "InspectInt"


class ValidationMsg(str, enum.Enum):
    """ Validation Message Type """
    Warning = "WarningMsg"
    Error = "ErrorMsg"


class AssetType(str, enum.Enum):

    Mortgage = "Mortgage"
    Loan = "Loan"
    Lease = "Lease"
    Receivable = "Receivable"
    Installment = "Installment"
    FixedAsset = "FixedAsset"


class SubAssetType(str, enum.Enum):

    Mortgage = "Mortgage"
    AdjustableRateMortgage = "AdjustableRateMortgage"
    ScheduleMortgageFlow = "ScheduleMortgageFlow"
    PersonalLoan = "PersonalLoan"
    RegularLease = "RegularLease"
    StepUpLease = "StepUpLease"
    Invoice = "Invoice"
    Installment = "Installment"
    FixedAsset = "FixedAsset"

dealStatInt = set(["BondPaidPeriod","PoolCollectedPeriod"])
dealStatBool = set([])
dealStatBalance = set([])
dealStatRate = set([])

inf = 179769313486231590772930519078902473361797697894230657273430081157732675805500963132708477322407536021120113879871393357658789768814416622492847430639474124377767893424865485276302219601246094119453082952085005768838150682342462881473913110540827237163350510684586298239947245938479716304835356329624224137216



import re
import dateparser
import toolz as tz
from lenses import lens

def rmDigitsInLine(x):
    return re.sub("\n\d+\n","",x)

def rmHeader(x):
    h1 = "归集日期 期初剩余本金 回收本金 回收利息 期末剩余本金"
    return re.sub(h1,"",x)

def removeComma(x):
    return x.replace(",","")

def hyphenToZero(x):
    return x.replace("-","0")

def removeEmpty(x):
    return [ _ for _ in x if _]

def removePg(xs):
    return [ x for x in xs if not re.match(r"^\d+$",x)]


def splitSpace(xs:list):
    return [ _.split(" ") for _ in xs]

def normalizedDate(x:str):
    return dateparser.parse(x).strftime("%Y-%m-%d")

def removeIfAllinSet(xs:list,s:set):
    return [ x for x in xs if not set(x) == s]


import pandas as pd
from lenses import lens


def compDf(x:pd.DataFrame, y:pd.DataFrame, names) -> pd.DataFrame:
    """ Compare two dataframe ,return the dataframe with difference """ 

    (nameA,nameB) = names
    x, y = x.align(y, fill_value=pd.NA)
    cmp = x.compare(y, result_names=(nameA,nameB))
    result = cmp.copy()
    for column in cmp.columns.levels[0]:
        if pd.api.types.is_numeric_dtype(cmp[column][nameA]) and pd.api.types.is_numeric_dtype(cmp[column][nameB]):
            result[column, 'diff'] = cmp[column][nameA].fillna(0) - cmp[column][nameB].fillna(0)
    return result.reindex(axis=1, level=1, labels=[nameA, nameB, 'diff']).sort_index(axis=1,level=0)


def compResult(r1:dict, r2:dict, names=("Left", "Right")):
    """ compare first to second result

    :param r1:single run result 1 (with read=True)
    :type r1: dict
    :param r2: single run result 2 (to be compared with r1)
    :type r2: dict
    :param names: name of the results, defaults to ("Left", "Right")
    :type names: tuple, optional
    """
    def cmpMap(x:dict, y:dict) -> dict:
        """ compare two map with the same keys """
        assert set(x.keys())==set(y.keys()), f"keys not match {x.keys()} vs {y.keys()}"
        cmp = {}
        for k,v in x.items():
            assert isinstance(v, pd.DataFrame), f"expecting DataFrame, got {type(v)}"
            assert isinstance(y[k], pd.DataFrame), f"expecting DataFrame, got {type(y[k])}"
            if not v.equals(y[k]):
                cmp[k] = compDf(v, y[k], names)
            else:
                cmp[k] = pd.DataFrame()
        return cmp

    # r1 -> Left
    # r2 -> Right
    comp_result = {}
    # pool check
    comp_result['pools'] = cmpMap(r1['pool']['flow'], r2['pool']['flow'])

    # expense check  
    comp_result['fees'] = cmpMap(r1['fees'], r2['fees'])

    # bond check
    if all([ isinstance(_, pd.DataFrame) for _ in r1['bonds'].values()]) and all([ isinstance(_, pd.DataFrame) for _ in r2['bonds'].values()]):
        comp_result['bonds'] = cmpMap(r1['bonds'], r2['bonds'])
    else:
        comp_result['bonds'] = {}
        for k,v in r1['bonds'].items():
            if isinstance(v, pd.DataFrame):
                comp_result['bonds'][k] = compDf(v, r2['bonds'][k])
            elif isinstance(v, dict):
                assert isinstance(r2['bonds'][k], dict), f"expecting dict, got {type(r2['bonds'][k])}"
                comp_result['bonds'][k] = cmpMap(v, r2['bonds'][k])

    # account check
    comp_result['accounts'] = cmpMap(r1['accounts'], r2['accounts'])

    # ledger check 
    comp_result['ledgers'] = cmpMap(r1['ledgers'], r2['ledgers']) if 'ledgers' in r1 else {}
    
    if 'triggers' in r1 and r1['triggers'] is not None:
        comp_result['triggers'] = {}
        for locName, tMap in r1['triggers'].items() :
            if len(tMap) == 0:
                continue
            comp_result['triggers'][locName] = cmpMap(r1['triggers'][locName], r2['triggers'][locName])

    return comp_result


def compTwoEngine(xEngine, yEngine, d, pAssump, rAssump):
    rx = xEngine.run(d, read=True, poolAssump=pAssump, runAssump=rAssump)
    ry = xEngine.run(d, read=True, poolAssump=pAssump, runAssump=rAssump)
    return compResult(rx, ry)





from rich.console import Console
from lenses import lens
import toolz as tz
from itertools import product
import dataclasses

#from .local.generic import Generic
from .local.component import * 
from .local.util import * 
from .local.generic import Generic

console = Console()

def mkDeal(x:dict, preCheck=True):
    name = getValWithKs(x, ['name', "名称", "comment", '备注'], defaultReturn="")
    
    dates = getValWithKs(x, ['dates', "date", "日期"])
    assert dates is not None, f"dates shouldn't be None"
    
    accs = list((accName,acc) for accName,acc in getValWithKs(x, ['accounts', "account", "账户", "帐户"]).items())
    assert accs is not None, f"accounts shouldn't be None"
    
    fees = list((fn,fv|{"name":fn,"名称":fn})
                for (fn,fv) in getValWithKs(x
                                            , ['expense', 'expenses', 'fee', "fees", "费用"]
                                            , defaultReturn={}).items())
    
    pool = getValWithKs(x, ['pool', "collateral", "资产池"])

    bonds = list((bn,bo)
                for bn,bo in getValWithKs(x, ['bond', "bonds", "notes", "债券", "支持证券"]).items())
    assert bonds is not None, f"bonds shouldn't be None"

    waterfall = getValWithKs(x, ['waterfall', "现金流分配", "分配规则"])
    assert waterfall is not None, f"waterfall shouldn't be None"
    
    collection = getValWithKs(x,['collect','colleciton', 'collectionRule', 'aggregation', '归集规则', '归集'])
    assert collection is not None , f"collection shouldn't be None"

    liqFacility = getValWithKs(x, ['liqFacility', 'liqProvider', "insurance"
                                    ,"cashAdvance", "liquidity", "流动性支持"
                                    , "流动性提供方", "增信方", "担保"], defaultReturn=None)
    
    rateSwap = getValWithKs(x, ['rateSwap','rateSwaps', "IRSwap", "利率互换"])
    
    rateCap = getValWithKs(x, ['rateCap',"rateCaps", "IRCap", "利率上限互换"])
    
    currencySwap = None  #TODO 

    trigger = getValWithKs(x, ['trigger', "triggers", "事件", "触发事件"], defaultReturn=None)
    
    status = getValWithKs(x, ['status', 'stage', "状态", "阶段"], defaultReturn="Amortizing")
    
    custom = getValWithKs(x, ['custom', "自定义"])
    
    ledgers = getValWithKs(x, ['ledger', "ledgers", "科目"])
    
    deal = Generic(
        name
        ,dates 
        ,pool
        ,accs
        ,bonds
        ,fees
        ,waterfall
        ,collection
        ,liqFacility
        ,rateSwap
        ,currencySwap
        ,trigger
        ,status
        ,None
        ,ledgers
        ,rateCap
    )

    return deal


def mkDealsBy(d, m: dict)->dict:
    "Input a deal, permunations, lenses ,and return a list of deals with variety"
    return {k: dataclasses.replace(d, **v) for k, v in m.items()} 


def setDealsBy(d, *receipes: list, init=None, **kwargs):
    "input a deal object, a list of tweaks( path, values) ,return an updated deal"
    if init:
        receipes = [(init & _[0], _[1]) for _ in receipes]
    for (p, v) in receipes:
        d &= p.set(v)
    return d


def prodDealsBy(d, *receipes, **kwargs) -> dict:
    inflated = [[(p, _) for _ in vs] for (p, vs) in receipes]
    permus = product(*inflated)
    if kwargs.get('guessKey', False) == True:
        return {strFromPath(v): setDealsBy(d, *v, **kwargs) for v in permus}
    return {v: setDealsBy(d, *v, **kwargs) for v in permus}


def setAssumpsBy(a, *receipes: list, init=None):
    if init:
        receipes = [(init & _[0], _[1]) for _ in receipes]
    for (p, v) in receipes:
        a &= p.set(v)
    return a


def prodAssumpsBy(a, *receipes, **kwargs):
    inflated = [[(p, _) for _ in vs] for (p, vs) in receipes]
    permus = product(*inflated)
    return {str(v): setAssumpsBy(a, *v, **kwargs) for v in permus}






from ..local.generic import Generic

defautlRateTrigger = {"defaultRateTrigger": 
                        {"condition":[("cumPoolDefaultedRate",),">", 0.05]
                        ,"effects":("newStatus","Accelerated")
                        ,"status":False
                        ,"curable":False}}


trigger01 = Generic(
    "Trigger_Default:01"
    ,{"cutoff":"2021-03-01","closing":"2021-04-01","firstPay":"2021-06-20"
     ,"payFreq":["DayOfMonth",20],"poolFreq":"MonthFirst","stated":"2030-01-01"}
    ,{'assets':[["Mortgage"
        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":20
          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
          ,{"currentBalance":2200
          ,"currentRate":0.08
          ,"remainTerm":20
          ,"status":"current"}]]}
    ,(("acc01",{"balance":0}),)
    ,(("A1",{"balance":1000
             ,"rate":0.07
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.08}
             ,"bondType":{"Sequential":None}})
      ,("B",{"balance":1000
             ,"rate":0.0
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.00}
             ,"bondType":{"Equity":None}
             }))
    ,tuple()
    ,{"amortizing":[
         ["accrueAndPayInt","acc01",["A1"]]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payIntResidual","acc01","B"]
     ]}
    ,[["CollectedInterest","acc01"]
      ,["CollectedPrincipal","acc01"]
      ,["CollectedPrepayment","acc01"]
      ,["CollectedRecoveries","acc01"]]
    ,None
    ,None
    ,None
    ,{"AfterCollect": defautlRateTrigger}
    ,("PreClosing","Amortizing")
    )

trigger02 = Generic(
    "Trigger_ongoing_Default:02"
    ,{"collect":["2021-10-01","2021-11-01"]
      ,"pay":["2021-09-20","2021-10-20"]
      ,"payFreq":["DayOfMonth",20],"poolFreq":"MonthFirst","stated":"2030-01-01"}
    ,{'assets':[["Mortgage"
        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":20
          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
          ,{"currentBalance":2200
          ,"currentRate":0.08
          ,"remainTerm":20
          ,"status":"current"}]]
      ,'issuanceStat':{"HistoryDefaults":50,"IssuanceBalance":1500}}
    ,(("acc01",{"balance":0}),)
    ,(("A1",{"balance":1000
             ,"rate":0.07
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.08}
             ,"bondType":{"Sequential":None}})
      ,("B",{"balance":1000
             ,"rate":0.0
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.00}
             ,"bondType":{"Equity":None}
             }))
    ,tuple()
    ,{"amortizing":[
         ["accrueAndPayInt","acc01",["A1"]]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payIntResidual","acc01","B"]
     ]}
    ,[["CollectedInterest","acc01"]
      ,["CollectedPrincipal","acc01"]
      ,["CollectedPrepayment","acc01"]
      ,["CollectedRecoveries","acc01"]]
    ,None
    ,None
    ,None
    ,{"AfterCollect": defautlRateTrigger}
    ,"Amortizing"
    )




defautlRollingRateTrigger = {"defaultRateTrigger": 
                              {"condition":[("avg"
                                              , ("cumPoolDefaultedRate",)
                                              , ("cumPoolDefaultedRateTill",-1,)
                                              , ("cumPoolDefaultedRateTill",-2,)
                                              ),">", 0.05]
                              ,"effects":("newStatus","Accelerated")
                              ,"status":False
                              ,"curable":False}}



trigger03 = Generic(
    "Trigger_rolling_Default rate:03"
    ,{"collect":["2021-10-01","2021-11-01"]
      ,"pay":["2021-09-20","2021-10-20"]
      ,"payFreq":["DayOfMonth",20],"poolFreq":"MonthFirst","stated":"2030-01-01"}
    ,{'assets':[["Mortgage"
        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":20
          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
          ,{"currentBalance":2200
          ,"currentRate":0.08
          ,"remainTerm":20
          ,"status":"current"}]]
      ,'issuanceStat':{"HistoryDefaults":50,"IssuanceBalance":1500}}
    ,(("acc01",{"balance":0}),)
    ,(("A1",{"balance":1000
             ,"rate":0.07
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.08}
             ,"bondType":{"Sequential":None}})
      ,("B",{"balance":1000
             ,"rate":0.0
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.00}
             ,"bondType":{"Equity":None}
             }))
    ,tuple()
    ,{"amortizing":[
         ["accrueAndPayInt","acc01",["A1"]]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payIntResidual","acc01","B"]
     ]}
    ,[["CollectedInterest","acc01"]
      ,["CollectedPrincipal","acc01"]
      ,["CollectedPrepayment","acc01"]
      ,["CollectedRecoveries","acc01"]]
    ,None
    ,None
    ,None
    ,{"AfterCollect": defautlRollingRateTrigger}
    ,"Amortizing"
    )


trigger04 = Generic(
    "Trigger with waterfall action"
    ,{"cutoff":"2021-03-01","closing":"2021-04-01","firstPay":"2021-06-20"
     ,"payFreq":["DayOfMonth",20],"poolFreq":"MonthFirst","stated":"2030-01-01"}
    ,{'assets':[["Mortgage"
        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":20
          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
          ,{"currentBalance":2200
          ,"currentRate":0.08
          ,"remainTerm":20
          ,"status":"current"}]]}
    ,(("acc01",{"balance":0}),)
    ,(("A1",{"balance":1000
             ,"rate":0.07
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.08}
             ,"bondType":{"Sequential":None}})
      ,("B",{"balance":1000
             ,"rate":0.0
             ,"originBalance":1000
             ,"originRate":0.07
             ,"startDate":"2020-01-03"
             ,"rateType":{"Fixed":0.00}
             ,"bondType":{"Equity":None}
             }))
    ,tuple()
    ,{"amortizing":[
         ["accrueAndPayInt","acc01",["A1"]]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payIntResidual","acc01","B"]
     ]}
    ,[["CollectedInterest","acc01"]
      ,["CollectedPrincipal","acc01"]
      ,["CollectedPrepayment","acc01"]
      ,["CollectedRecoveries","acc01"]]
    ,None
    ,None
    ,None
    ,{
        "AfterCollect":
          {"poolDef":
            {"condition":[("cumPoolDefaultedBalance",),">",20]
            ,"effects":("actions"
                        ,["calcInt", "A1"]
                        ,["payInt","acc01",["A1"]]
                        )
            ,"status":False
            ,"curable":False}
          }}
    ,("PreClosing","Amortizing")
    )


import matplotlib.pyplot as plt
from matplotlib import font_manager
from functools import reduce
import numpy as np
import pandas as pd
import toolz as tz
import logging
from .base import china_cumStats, english_cumStats,english_non_balance_flow,china_non_balance_flow
from .util import guess_locale
from .cf import readFlowsByScenarios

"""
 chinese character issue.

 https://github.com/Clarmy/mplfonts
 
 from mplfonts import use_font
 use_font('Noto Serif CJK SC')

    
"""

dmap = {
    "cn":{
        "bond":"债券","scenario":"场景"
    },
    "en":{
        "bond":"bond","scenario":"scenario"
    }
}


def init_plot_fonts():
    define_list = ['Source Han Serif CN', 'Microsoft Yahei', 'STXihei','Heiti TC']
    support_list = font_manager.findSystemFonts(fontpaths=None, fontext='ttf')
    font_p = font_manager.FontProperties()
    try:
        for sl in support_list:
            f = font_manager.get_font(sl)
            if f.family_name in set(define_list):
                font_p.set_family(f.family_name)
                font_p.set_size(14)
                return font_p
    except RuntimeError as e:
        logging.error("中文字体载入失败")
        return None

font_p = init_plot_fonts()

def plotPool(p:pd.DataFrame):
    """ Plot the pool flows """

    dates = p.index.to_list()

    #remove cumut stats
    fieldsToRemove = set(china_cumStats+english_cumStats+china_non_balance_flow+english_non_balance_flow)
    
    flows = tz.pipe(p.fillna(0)
                    ,p.to_dict(orient='list')
                    ,lambda x: tz.keyfilter( lambda y: y not in fieldsToRemove , x)
    )


    fig, ax = plt.subplots()
    ax.stackplot(dates, flows.values(),
                labels= flows.keys(), alpha=0.8)
    ax.legend(loc='upper right', reverse=True)
    ax.set_title('Pool Flows')
    ax.set_xlabel('Date')
    ax.set_ylabel('Balance')

    plt.show()


def plotFlowByScenarios(rs:dict, path):
    """ Plot time series balance on return of multi-scenario/multi-structure """

    df = readFlowsByScenarios(rs, path)
    
    return df


def plot_bond(rs, bnd, flow='本息合计'):
    """Plot bonds across scenarios"""
    plt.figure(figsize=(12,8))
    _alpha = 0.8
    locale = guess_locale(list(rs.values())[0])
    for idx, s in enumerate(rs):
        plt.step(s['bonds'][bnd].index,s['bonds'][bnd][[flow]], alpha=_alpha, linewidth=5, label=f"{dmap[locale]['scenario']}-{idx}")

    plt.legend(loc='upper left', prop=font_p)
    plt.title(f"{len(rs)} {dmap[locale]['scenario']} {dmap[locale]['bond']}:{bnd} - {flow}", fontproperties=font_p)

    plt.grid(True)
    plt.axis('tight')
    plt.xticks(rotation=30)

    current_values = plt.gca().get_yticks()
    plt.gca().set_yticklabels(['{:.0f}(w)'.format(x/10000) for x in current_values])
    return plt


def plot_bonds(r, bnds:list, flow='本息合计'):
    "Plot bond flows with in a single run"

    locale = guess_locale(r)
    plt.figure(figsize=(12,8))
    _alpha =  0.8

    for b in bnds:
        b_flow = r['bonds'][b]
        plt.step(b_flow.index,b_flow[[flow]], alpha=_alpha, linewidth=5, label=f"{dmap[locale]['bond']}-{b}")

    plt.legend(loc='upper left', prop=font_p)
    bnd_title = ','.join(bnds)
    plt.title(f"{dmap[locale]['bond']}:{bnd_title} - {flow}", fontproperties=font_p)

    plt.grid(True)
    plt.axis('tight')
    plt.xticks(rotation=30)

    current_values = plt.gca().get_yticks()
    plt.gca().set_yticklabels(['{:.0f}(w)'.format(x/10000) for x in current_values])
    return plt


# def plot_by_scenario(rs, flowtype, flowpath):
#    "Plot with multiple scenario"
#    plt.figure(figsize=(12,8))
#    scenario_names = rs.keys()
#    dflows = [query(rs,[s]+flowpath) for s in scenario_names]
#    _alpha =  0.8
#
#    x_labels = reduce(lambda acc,x:acc.union(x) ,[ _.index for _ in dflows ]).unique()
#    x = np.arange(len(x_labels))
#    width = 1 
#    step_length = width / (len(scenario_names)+1)
#
#    for (idx,(scen,dflow)) in enumerate(zip(scenario_names,dflows)):
#        if flowtype=="balance":
#            cb = consolStmtByDate(dflow)
#            plt.step(cb.index, cb, alpha=_alpha, linewidth=5, label=f"{scen}")
#        elif flowtype=="amount":
#            cb = aggStmtByDate(dflow)
#            _bar = plt.bar(x+idx*step_length,cb,width=step_length,label=scen)
#        else:
#            plt.plot(dflow.index,dflow, alpha=_alpha, linewidth=5, label=f"{scen}")
#
#    plt.legend(scenario_names,loc='upper right', prop=font_p)
#    plt.grid(True)
#    plt.axis('tight')
#    plt.xticks(ticks=x,labels=x_labels,rotation=30)


def plot_bs(x, excludeItems=["FeeDue","IntAccrue","Account"]):
    dates = x.index.to_list() 

    liabilityItems = x['liability'].to_dict(orient='list')
    assetItems = x['asset'].to_dict(orient='list')
    
    highest_y1 = max([max(v) for k, v in liabilityItems.items()])
    highest_y2 = max([max(v) for k, v in assetItems.items()])
    
    if "FeeDue" in set(excludeItems):
        liabilityItems = {k: v for (k, v) in liabilityItems.items() if (not k.startswith("Fee Due:"))}

    if "IntAccrue" in set(excludeItems):
        liabilityItems = {k: v for (k, v) in liabilityItems.items() if (not k.startswith("Accured Int:")) }
   
    if "Account" in set(excludeItems):
        assetItems = {k: v for (k, v) in assetItems.items() if k not in excludeItems }

    balanceItems = {
        'Asset': assetItems,
        'Liability': liabilityItems,
    }

    x = np.arange(len(dates))  # the label locations
    width = 0.40  # the width of the bars
    multiplier = 0

    fig, ax = plt.subplots(layout='constrained')

    for assetClass, assetBreakdown in balanceItems.items():
        offset = width * multiplier
        bottom = np.zeros(len(dates))
        for (breakdownsK, breakdowns) in assetBreakdown.items():
            rects = ax.bar(x + offset, breakdowns, width, label=breakdownsK, bottom=bottom)
            bottom += breakdowns

            ax.bar_label(rects, padding=1, label_type="center")
        multiplier += 1

    ax.set_ylabel('Amount')
    ax.set_title('Projected Capital Structure')
    ax.set_xticks(x + width/2, dates)
    ax.legend(loc='upper right')
    ax.set_ylim(0, max([highest_y1, highest_y2])*1.2)
    plt.show()


from collections import namedtuple
from typing import List, Tuple


def mkTag(x: tuple | str) -> dict:
    match x:
        case (tagName, tagValue):
            return {"tag": tagName, "contents": tagValue}
        case (tagName):
            return {"tag": tagName}

def readAeson(x:dict):
    match x:
        case {"tag": tag,"contents": contents} if isinstance(contents,list):
            return {tag: [readAeson(c) for c in contents]}
        case {"tag": tag,"contents": {'numerator': n,'denominator': de}} :
            return {tag: (n/de)}
        case {"tag": tag,"contents": contents} if isinstance(contents,dict):
            return {tag: {k: readAeson(v) for k,v in contents.items()} }            
        case {"tag": tag,"contents": contents} if isinstance(contents,str):
            return {tag: contents}   
        case {"tag": tag, **kwargs} if len(kwargs)>1 :
            return {k: readAeson(v) for k,v in kwargs.items()} | {"tag": tag}
        case {"tag": tag} if isinstance(tag,str):
            return tag   
        case {'numerator': n,'denominator': de}:
            return (n/de)
        case x if isinstance(x,dict):
            return {k:readAeson(_x) for k,_x in x.items()}
        case x if isinstance(x,list):
            return [readAeson(_x) for _x in x]
        case n if isinstance(n,float) or isinstance(n,int):
            return n
        case m if isinstance(m, dict):
            return {k: readAeson(v) for k,v in m.items() }
        case s if isinstance(s, str):
            return s
        case None:
            return None
        case {}:
            return {}
        case s if isinstance(s,str):
            return s
        case _:
            raise RuntimeError("failed to match",x)





from .local.interface import mkTag
from .validation import vStr,vBool,vNum
from .local.component import mkDs

def mkTweak(x):
    match x:
        case "stressDefault":
            return mkTag(("StressPoolDefault", [1.0, 500]))
        case ("stressDefault",l,h):
            return mkTag(("StressPoolDefault", [l, h]))
        case "stressPrepayment":
            return mkTag(("StressPoolPrepayment",  [1.0, 500]))
        case ("stressPrepayment", l, h):
            return mkTag(("StressPoolPrepayment", [l, h]))
        case ("maxSpread", bn):
            return mkTag(("MaxSpreadTo", [ vStr(bn), (1.0, 500)]))
        case ("maxSpread", bn, l, h):
            return mkTag(("MaxSpreadTo", [ vStr(bn), (l, h)]))
        case ("splitBalance", bn1 ,bn2):
            return mkTag(("SplitFixedBalance", [vStr(bn1), vStr(bn2), (1.0, 500)]))
        case ("splitBalance", bn1, bn2, l, h):
            return mkTag(("SplitFixedBalance", [vStr(bn1), vStr(bn2), (l, h)]))
        case _:
            raise RuntimeError(f"failed to match {x}:mkTweak")

def mkStop(x):
    match x:
        case ("bondIncurLoss", bn):
            return mkTag(("BondIncurLoss", vStr(bn)))
        case ("bondIncurPrinLoss", bn, amt):
            return mkTag(("BondIncurPrinLoss", [vStr(bn), vNum(amt)]))
        case ("bondIncurIntLoss", bn, amt):
            return mkTag(("BondIncurIntLoss", [vStr(bn), vNum(amt)]))
        case ("bondPricingEqOriginBal", bn, f1, f2):
            return mkTag(("BondPricingEqOriginBal", [vStr(bn), vBool(f1), vBool(f2)] ))
        case ("bondMetTargetIrr", bn, irr):
            return mkTag(("BondMetTargetIrr", [vStr(bn), vNum(irr)]))
        case ("byFormula", ds, target):
            return mkTag(("BalanceFormula", [mkDs(ds), vNum(target)]))
        case _:
            raise RuntimeError(f"failed to match {x}:mkStop")



import json, getpass, enum, os, pickle
from importlib.metadata import version
from json.decoder import JSONDecodeError
from dataclasses import dataclass
import requests
from rich.console import Console
import toolz as tz
from functools import partial
from lenses import lens

from requests.exceptions import ConnectionError, ReadTimeout

from .validation import isValidUrl
from .local.util import mapValsBy \
                              , _read_cf, _read_asset_pricing, mergeStrWithDict \
                              , earlyReturnNone, searchByFst, filter_by_tags \
                              , enumVals, lmap, inferPoolTypeFromAst, getValWithKs, mapNone\
                              
from .local.component import mkPool, mkAssumpType, mkNonPerfAssumps, mkLiqMethod \
                                   , mkAssetUnion, mkRateAssumption, mkDatePattern, mkPoolType

from .local.base import ValidationMsg
from .exception import AbsboxError, VersionMismatch, EngineError
from .local.interface import mkTag, readAeson
from .rootFinder import mkTweak, mkStop

VERSION_NUM = version("absbox")
console = Console()

__all__ = ["API", "Endpoints", "RunReqType", "RunResp" , "EnginePath"]


class Endpoints(str, enum.Enum):
    """API endpoints for engine server

    """
    RunAsset = "runAsset"
    """Run single asset endpoint"""
    RunPool = "runPool"
    """Run a pool of asset endpoint"""
    RunPoolByScenarios = "runPoolByScenarios"
    """Run a pool of asset with multiple scenarios endpoint"""
    RunDeal = "runDeal"
    """Run a single deal endpoint"""
    RunDealByScnearios = "runDealByScenarios"
    """Run a single deal with multiple scenarios endpoint"""
    RunMultiDeal = "runMultiDeals"
    """Run multiple deals endpoint"""
    RunDealByRunScenarios = "runDealByRunScenarios"
    """Run a single deal with multiple deal run scenarios endpoint"""
    RunByCombo = "runByCombo"
    """Run mulitple sensitivities"""
    RunRootFinder = "runByRootFinder"
    """Run root finder"""
    RunDate = "runDate"
    """Run Dates from a datepattern """
    Version = "version"
    """Get version of engine server endpoint"""


class RunReqType(str, enum.Enum):
    """Request run type
    """
    Single = "SingleRunReq"
    """ Single Deal With A Single Assumption """
    MultiScenarios = "MultiScenarioRunReq"
    """ Single Deal With Multiple Assumptions """
    MultiRunScenarios = "MultiRunAssumpReq"
    """ Single Deal With Multiple Assumptions """
    MultiStructs = "MultiDealRunReq"
    """ Multiple Deals With Single Assumption """
    SinglePool = "SingleRunPoolReq"
    """ Single Pool With Single Assumption """
    MultiPoolScenarios = "MultiScenarioRunPoolReq"
    """ Single Pool With Multiple Assumptions """
    ComboSensitivity = "MultiComboReq"
    """ Multiple sensitivities """
    RootFinder = "RootFinderReq"
    """ Root Finder Run """


class RunResp(int, enum.Enum):
    """Internal
    """
    DealResp = 0
    PoolResp = 1
    LogResp = 2


class EnginePath(str, enum.Enum):
    """ Enum class representing shortcut engine paths for the client.
    """
    LOCAL = "http://localhost:8081"
    """ Use local default server """
    PROD = "https://absbox.org/api/latest"
    DEV = "https://absbox.org/api/dev"
    NY_DEV = "https://spv.run/api/dev"
    NY_PROD = "https://spv.run/api/latest"
    LDN_DEV = "https://ldn.spv.run/api/dev"
    LDN_PROD = "https://ldn.spv.run/api/latest"
    USE_ENV = "USE_ENV"
    """ USE_ENV (str): Use environment variable (`ABSBOX_SERVER`) for engine path """


def PickApiFrom(Apilist:list, **kwargs):
    """ Auto init API instance from a list of API urls with version check

    :param Apilist: list of API urls
    :type Apilist: list
    """
    def pingApi(x):
        try:
            r = requests.get(f"{x.value}/{Endpoints.Version.value}", verify=False, timeout=5 ,headers={"Origin":"http://localhost:8001"}).text 
            return json.loads(r) 
        except Exception as e:
            return ("Error",e)

    _, libVersion, _ = VERSION_NUM.split(".")

    apiResps = [{"url":api,"resp":pingApi(api)} for api in Apilist ]
    validApis = tz.pipe(apiResps
                    ,lambda apis: list(filter(lambda x:"Error" not in x['resp'],apis))
                    ,lambda apis: lens.Each()['resp']['_version'].modify(lambda x: x.split("."))(apis)
                    ,lambda apis: filter(lambda x:x['resp']['_version'][1]==libVersion, apis) 
                    ,lambda apis: sorted(apis,key=lambda x: x['resp']['_version'][2])
    )

    r = list(validApis)

    if len(r)>0: 
        return API(r[0],**kwargs)
    else:
        raise AbsboxError(f"❌ No valid API found in list match current lib version {libVersion}, from list:{apiResps}")


@dataclass
class API:
    """ API to connect to engine server, handling requests and responses 

    :return: API instance
    :rtype: API
    """

    url: str
    """url of engine server"""
    lang: str = "english"
    """language of response from server, defaults to 'english' """
    check: bool = True
    """ flag to ensure version match between client and server """
    server_info = {}
    """ internal """
    version = VERSION_NUM.split(".")
    """ internal """
    hdrs = {'Content-type': 'application/json', 'Accept': '*/*', 'Accept-Encoding': 'gzip'}
    """ internal """
    session = None
    """ internal """
    debug = False
    """ internal """

    def __post_init__(self) -> None:
        """Init the API instance with url and perform version check

        :raises ConnectionRefusedError: Failed to connect to server
        :raises RuntimeError: Failed to get version info from server
        :raises VersionMismatch: Failed to match version between client and server
        """
        if self.url == EnginePath.USE_ENV.value:
            urlInEnv = os.environ.get("ABSBOX_SERVER")
            if urlInEnv is None:
                raise AbsboxError(f"❌No ABSBOX_SERVER found in environment variable")
            else:
                self.url = isValidUrl(urlInEnv).rstrip("/")
        else:
            self.url = isValidUrl(self.url).rstrip("/")

        console.print(f"Connecting engine server -> {self.url}")

        if self.lang not in ["chinese", "english"]:
            raise AbsboxError(f"❌Invalid language:{self.lang}, only support 'chinese' or 'english' ")

        try:
            _r = requests.get(f"{self.url}/{Endpoints.Version.value}", verify=False, timeout=5, headers = {"Origin":"http://localhost:8001"}).text
        except (ConnectionRefusedError, ConnectionError):
            raise AbsboxError(f"❌Error: Can't not connect to API server {self.url}")
        if _r is None:
            raise RuntimeError(f"Failed to get version from url:{self.url}")
        self.server_info = self.server_info | json.loads(_r)
        engine_version = self.server_info['_version'].split(".")
        if self.check and (self.version[1] != engine_version[1]):
            console.print("pls upgrade your api package by: pip -U absbox")
            raise VersionMismatch('.'.join(self.version), '.'.join(engine_version))
        self.session = requests.Session()
        console.print(f"✅Connected, local lib:{'.'.join(self.version)}, server:{'.'.join(engine_version)}")

    def build_run_deal_req(self, run_type, deal, perfAssump=None, nonPerfAssump=[], rtn = []) -> str:
        """build run deal requests: (single run, multi-scenario run, multi-struct run)

        :meta private:
        :param run_type: type of run request, RunReqType.Single.value, RunReqType.MultiScenarios.value, RunReqType.MultiStructs.value
        :type run_type: str
        :param deal: a deal/transaction object
        :type deal: _type_
        :param perfAssump: a tuple of pool level assumption(Default/Prepayment/Recovery) for single run. a map for multi-scenario run, defaults to None
        :type perfAssump: _type_, optional
        :param nonPerfAssump: a list of deal level assumptions, defaults to []
        :type nonPerfAssump: list, optional
        :raises RuntimeError: _description_
        :return: a string of request body to be sent out to engine
        :rtype: str

        """
        r = None

        match run_type:
            case "Single" | "S":
                _deal = deal.json if hasattr(deal, "json") else deal
                _perfAssump = earlyReturnNone(mkAssumpType, perfAssump)
                _nonPerfAssump = mkNonPerfAssumps({}, nonPerfAssump)
                r = mkTag((RunReqType.Single.value, [rtn, _deal, _perfAssump, _nonPerfAssump]))
            case "MultiScenarios" | "MS":
                _nonPerfAssump = mkNonPerfAssumps({}, nonPerfAssump)
                _deal = deal.json if hasattr(deal, "json") else deal
                mAssump = mapValsBy(perfAssump, mkAssumpType)
                r = mkTag((RunReqType.MultiScenarios.value, [rtn, _deal, mAssump, _nonPerfAssump]))
            case "MultiStructs" | "MD" :
                _nonPerfAssump = mkNonPerfAssumps({}, nonPerfAssump)
                mDeal = {k: v.json if hasattr(v, "json") else v for k, v in deal.items()}
                _perfAssump = mkAssumpType(perfAssump)
                r = mkTag((RunReqType.MultiStructs.value, [rtn, mDeal, _perfAssump, _nonPerfAssump]))
            case "MultiRunScenarios" | "MRS" if isinstance(nonPerfAssump,dict):
                _deal = deal.json if hasattr(deal, "json") else deal
                _perfAssump = earlyReturnNone(mkAssumpType, perfAssump)
                mRunAssump = mapValsBy(nonPerfAssump, lambda x: mkNonPerfAssumps({}, x))
                r = mkTag((RunReqType.MultiRunScenarios.value, [rtn, _deal, _perfAssump, mRunAssump]))
            case "ComboSensitivity" | "CS" if isinstance(nonPerfAssump,dict) and isinstance(perfAssump,dict) and isinstance(deal,dict):
                mDeal = {k: v.json if hasattr(v, "json") else v for k, v in deal.items()}
                mAssump = mapValsBy(perfAssump, mkAssumpType)
                if mAssump == {}:
                    mAssump = {"Empty": None}
                mRunAssump = mapValsBy(nonPerfAssump, lambda x: mkNonPerfAssumps({}, x))
                if mRunAssump == {}:
                    mRunAssump = {"Empty":{}}
                r = mkTag((RunReqType.ComboSensitivity.value, [rtn, mDeal, mAssump, mRunAssump]))
            case ("RootFinder", tweak, stop):
                _deal = deal.json if hasattr(deal, "json") else deal
                _perfAssump = earlyReturnNone(mkAssumpType, perfAssump)
                _nonPerfAssump = mkNonPerfAssumps({}, nonPerfAssump)
                dealRunInput= (_deal, _perfAssump, _nonPerfAssump, [])
                r = mkTag((RunReqType.RootFinder.value, [dealRunInput, mkTweak(tweak), mkStop(stop)]))
            case ("FirstLoss", bn) | ("FL", bn):
                _deal = deal.json if hasattr(deal, "json") else deal
                _perfAssump = mkAssumpType(perfAssump)
                _nonPerfAssump = mkNonPerfAssumps({}, nonPerfAssump)
                r = mkTag(("FirstLossReq", [(_deal, _perfAssump, _nonPerfAssump), bn]))
            case ("MaxSpreadToFaceReq", bn, bondFlag, feeFlag):
                _deal = deal.json if hasattr(deal, "json") else deal
                _perfAssump = mkAssumpType(perfAssump)
                _nonPerfAssump = mkNonPerfAssumps({}, nonPerfAssump)
                r = mkTag(("MaxSpreadToFaceReq", [(_deal, _perfAssump, _nonPerfAssump), bn, bondFlag, feeFlag]))
            case _:
                raise RuntimeError(f"Failed to match run type:{run_type}")
        
        assert r is not None, f"Failed to build request for run type:{run_type}"
        try:
            return json.dumps(r, ensure_ascii=False)
        except TypeError as e:
            raise AbsboxError(f"❌Failed to convert request to json:{e}")

        

    def build_pool_req(self, pool, poolAssump, rateAssumps, isMultiScenario=False, breakdown = False) -> str:
        """build pool run request: (single run, multi-scenario run)

        :meta private:
        :param pool: a pool object, could be a single pool or a pool map
        :type pool: _type_
        :param poolAssump: a tuple of pool level assumption(Default/Prepayment/Recovery) for single run. a map indicates multi-scenario run
        :type poolAssump: _type_
        :param rateAssumps: a list of rate assumptions
        :type rateAssumps: _type_
        :raises RuntimeError: _description_
        :return: a string of request body to be sent out to engine
        :rtype: str

        """
        r = None
        _rateAssump = lmap(mkRateAssumption, rateAssumps) if rateAssumps else None
        assetDate = getValWithKs(pool, ['cutoffDate', '封包日'])

        def buildPoolType(p) -> dict:
            """ build type for `PoolTypeWrap` """
            assetTag = inferPoolTypeFromAst(p) if (('assets' in p) or ('清单' in p)) else "UPool"
            _p = tz.dissoc(p, 'cutoffDate')
            if assetTag == "UPool":
                return mkTag((assetTag, mkPoolType(assetDate, _p, True)))
            else:
                return mkTag((assetTag, mkPoolType(assetDate, _p, False)))

        if not isMultiScenario:
            r = mkTag((RunReqType.SinglePool.value, [breakdown, buildPoolType(pool), mkAssumpType(poolAssump), _rateAssump]))
        else:
            r = mkTag((RunReqType.MultiPoolScenarios.value, [breakdown, buildPoolType(pool), mapValsBy(poolAssump, mkAssumpType), _rateAssump]))

        return json.dumps(r, ensure_ascii=False)

    @staticmethod    
    def _getWarningMsg(msgs, flag:bool) -> []:
        """get warning message from server response
        :return: list of warning messages
        :rtype: list
        """
        if flag:
            return [f"{x['contents']}" for x in filter_by_tags(msgs, enumVals(ValidationMsg))]
        else:
            return []

    def run(self, deal,
            poolAssump=None,
            runAssump=[],
            read=True,
            showWarning=True,
            rtn = [],
            debug=False) -> dict:
        """ run deal with pool and deal run assumptions

        :param deal: a deal object
        :type deal: Generic | SPV
        :param poolAssump: pool performance assumption, a tuple for single run/ a dict for multi-sceanrio run, defaults to None
        :type poolAssump: tuple, optional
        :param runAssump: deal level assumption, defaults to []
        :type runAssump: list, optional
        :param read: flag to convert result to pandas dataframe, defaults to True
        :type read: bool, optional
        :param showWarning: flag to show warnings, defaults to True
        :type showWarning: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        :return: result of run, a dict of dataframe if `read` is True.
        :rtype: dict

        """

        if (not isinstance(poolAssump, tuple)) and (poolAssump is not None):
            raise AbsboxError(f"❌ poolAssump should be a tuple but got {type(poolAssump)}")


        # if run req is a multi-scenario run
        url = f"{self.url}/{Endpoints.RunDeal.value}"

        # construct request
        req = self.build_run_deal_req("Single", deal, poolAssump, runAssump, rtn=rtn)
        if debug:
            return req

        result = self._send_req(req, url)

        if result is None or 'error' in result or 'Left' in result:
            leftVal = result.get("Left","")
            raise AbsboxError(f"❌ Failed to get response from run: {leftVal}")
        result = result['Right']

        if (wMsgs:=self._getWarningMsg(result[RunResp.LogResp.value], showWarning)):
            console.print("Warning Message from server:"+"\n".join(wMsgs))

        if read:
            return deal.read(result)
        else:
            return result

    def runByScenarios(self, deal,
                    poolAssump=None,
                    runAssump=[],
                    read=True,
                    showWarning=True,
                    debug=False) -> dict :
        """ run deal with multiple scenarios, return a map

        :param deal: _description_
        :type deal: _type_
        :param poolAssump: _description_, defaults to None
        :type poolAssump: dict, optional
        :param runAssump: _description_, defaults to []
        :type runAssump: list, optional
        :param read: if read response into dataframe, defaults to True
        :type read: bool, optional
        :param showWarning: if show warning messages from server, defaults to True
        :type showWarning: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        :return: a dict with scenario names as keys
        :rtype: dict        
        """

        url = f"{self.url}/{Endpoints.RunDealByScnearios.value}"
        req = self.build_run_deal_req("MultiScenarios", deal, poolAssump, runAssump)

        if debug:
            return req

        if runAssump is None or searchByFst(runAssump, "pricing") is None:
            result = self._send_req(req, url)
        else:
            result = self._send_req(req, url, timeout=30)

        if result is None or 'error' in result or "Left" in set(tz.concat([ _.keys() for _ in result.values()])):
            leftVal = { k:v['Left'] for k,v in result.items() if "Left" in v }
            raise AbsboxError(f"❌ Failed to get response from run: {leftVal}")
        
        result = tz.valmap(lambda x:x['Right'] ,result)

        rawWarnMsgByScen = {k: self._getWarningMsg(v[RunResp.LogResp.value],showWarning) for k, v in result.items()}
        for scen, msgs in rawWarnMsgByScen.items():
            if len(msgs)>0:
                console.print(f"Warning Message from server for {scen}:"+"\n".join(msgs))

        if read:
            return tz.valmap(deal.read, result)
        else:
            return result

    def read_single(self, breakdown, pool_resp) -> tuple:
        """ read pool run response from engine and convert to dataframe

        :param pool_resp: (pool raw cashflow, pool statistics)
        :type pool_resp: tuple
        :return: (pool Cashflow in dataFrame, pool statistics)
        :rtype: tuple
        """

        (pool_flow, pool_breakdown_flow) = pool_resp
        result = _read_cf(pool_flow['contents'][1], self.lang)
        if not breakdown:
            return {"flow":result}
        else:
            assert pool_breakdown_flow is not None, "Breakdown flow is None"
            assert len(pool_breakdown_flow)>0, "Breakdown flow is empty"
            return {"flow":result 
                    ,"breakdown": [ _read_cf(_['contents'][1], self.lang)
                                    for _ in pool_breakdown_flow  ]
                    }

    def runPoolByScenarios(self, pool, poolAssump, rateAssump=None, read=True, breakdown = False,debug=False) -> dict :
        """ run a pool with multiple scenario ,return result as map , with key same to pool assumption map

        :param pool: pool map
        :type pool: dict
        :param poolAssump: assumption map
        :type poolAssump: dict
        :param rateAssump: _description_, defaults to None
        :type rateAssump: _type_, optional
        :param read: if read response into dataframe, defaults to True
        :type read: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        :return: a dict with scenario names as keys
        :rtype: dict
        """

        url = f"{self.url}/{Endpoints.RunPoolByScenarios.value}"
        req = self.build_pool_req(pool, poolAssump, rateAssump, isMultiScenario=True)

        if debug:
            return req

        result = self._send_req(req, url)
        
        if result is None or 'error' in result or "Left" in set(tz.concat([ _.keys() for _ in result.values()])):
            leftVal = { k:v['Left'] for k,v in result.items() if "Left" in v }
            raise AbsboxError(f"❌ Failed to get response from run: {leftVal}")
        
        result = tz.valmap(lambda x:x['Right'] ,result)

        if read:
            return result & lens.Values().Values().modify(partial(self.read_single, breakdown))
        return result

    def runPool(self, pool, poolAssump=None, rateAssump=None, read=True, debug=False, breakdown = False, **kwargs) -> tuple:
        """perform pool run with pool and rate assumptions

        :param pool: a pool object
        :type pool: object
        :param poolAssump: pool performance assumption, a tuple for single run and a map for multi-scenario run, defaults to None
        :type poolAssump: tuple
        :param rateAssump: a list of interest rate assumptions, default to None
        :type rateAssump: tuple, optional
        :param read: flag to convert result to pandas dataframe, default to True
        :type read: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        :return: tuple of cashflow and pool statistics
        :rtype: tuple
        """

        if (not isinstance(poolAssump, tuple)) and (poolAssump is not None):
            raise AbsboxError(f"❌ poolAssump should be a tuple but got {type(poolAssump)}")

        url = f"{self.url}/{Endpoints.RunPool.value}"

        req = self.build_pool_req(pool, poolAssump, rateAssump, isMultiScenario=False, breakdown = breakdown)

        if debug:
            return req

        result = self._send_req(req, url, **kwargs)

        if result is None or 'error' in result or 'Left' in result:
            leftVal = result.get("Left","")
            raise AbsboxError(f"❌ Failed to get response from run: {leftVal}")

        result = result['Right']

        if read:
            return result & lens.Values().modify(partial(self.read_single, breakdown))
        else:
            return result

    def runStructs(self, deals, poolAssump=None, nonPoolAssump=None, runAssump=None, rtn=[], read=True, debug=False) -> dict:
        """run multiple deals with same assumption

        :param deals: a dict of deals
        :type deals: dict
        :param poolAssump: _description_, defaults to None
        :type poolAssump: _type_, optional
        :param nonPoolAssump: _description_, defaults to None
        :type nonPoolAssump: _type_, optional
        :param runAssump: _description_, defaults to None
        :type runAssump: _type_, optional
        :param read: _description_, defaults to True
        :type read: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        :return: a map of results
        :rtype: dict
        """
        assert isinstance(deals, dict), f"Deals should be a dict but got {type(deals)}"

        url = f"{self.url}/{Endpoints.RunMultiDeal.value}" 
        _poolAssump = mkAssumpType(poolAssump) if poolAssump else None 
        _nonPerfAssump = mkNonPerfAssumps({}, mapNone(nonPoolAssump,[]) + mapNone(runAssump,[]))
        req = json.dumps(mkTag(("MultiDealRunReq"
                                 ,[rtn
                                   ,tz.valmap(lambda x:x.json, deals)
                                   ,_poolAssump
                                   ,_nonPerfAssump]))
                         ,ensure_ascii=False)

        if debug:
            return req
        result = self._send_req(req, url)

        if result is None or 'error' in result or "Left" in set(tz.concat([ _.keys() for _ in result.values()])):
            leftVal = { k:v['Left'] for k,v in result.items() if "Left" in v }
            raise AbsboxError(f"❌ Failed to get response from run: {leftVal}")

        result = tz.valmap(lambda x:x['Right'] ,result)
        
        if read:
            return {k: deals[k].read(v) for k, v in result.items()}    
        else:
            return result

    def runByDealScenarios(self, deal,
                    poolAssump=None,
                    runAssump={},
                    read=True,
                    showWarning=True,
                    debug=False) -> dict :
        """ run deal with multiple run assumption, return a map

        :param deal: _description_
        :type deal: _type_
        :param poolAssump: _description_, defaults to None
        :type poolAssump: dict, optional
        :param runAssump: _description_, defaults to {}
        :type runAssump: dict
        :param read: if read response into dataframe, defaults to True
        :type read: bool, optional
        :param showWarning: if show warning messages from server, defaults to True
        :type showWarning: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        :return: a dict with scenario names as keys
        :rtype: dict        
        """

        url = f"{self.url}/{Endpoints.RunDealByRunScenarios.value}"
        req = self.build_run_deal_req("MRS", deal, poolAssump, runAssump)

        if debug:
            return req

        result = self._send_req(req, url, timeout=30)

        if result is None or 'error' in result or "Left" in set(tz.concat([ _.keys() for _ in result.values()])):
            leftVal = { k:v['Left'] for k,v in result.items() if "Left" in v }
            raise AbsboxError(f"❌ Failed to get response from run: {leftVal}")

        result = tz.valmap(lambda x:x['Right'] ,result)

        rawWarnMsgByScen = {k: self._getWarningMsg(v[RunResp.LogResp.value],showWarning) for k, v in result.items()}
        for scen, msgs in rawWarnMsgByScen.items():
            if len(msgs)>0:
                console.print(f"Warning Message from server for {scen}:"+"\n".join(msgs))

        if read:
            return tz.valmap(deal.read, result)
        else:
            return result

    def runByCombo(self, 
                    dealMap, 
                    poolAssump={}, 
                    runAssump={}, 
                    read=True, showWarning=True, debug=False) -> dict:
        """ run mulitple deals with multiple pool assumption/ with multiple run assumption, return a map 
        
        :param dealMap: a dict of deals
        :type dealMap: _type_
        :param poolAssump: a dict of pool assumptions, defaults to {}
        :type poolAssump: dict, optional
        :param runAssump: a dict of run assumptions, defaults to {}
        :type runAssump: dict, optional
        :param read: if read response into dataframe, defaults to True
        :type read: bool, optional
        :param showWarning: if show warning messages from server, defaults to True
        :type showWarning: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        """

        url = f"{self.url}/{Endpoints.RunByCombo.value}"

        if len(dealMap) == 0:
            raise AbsboxError(f"❌ No deals found in dealMap,at least one deal is required")

        if "^" in " ".join(tz.concatv([str(_) for _ in dealMap.keys()],[str(_) for _ in poolAssump.keys()],[str(_) for _ in runAssump.keys()])):
            raise AbsboxError(f"❌ Deal name should not contain '^' ")

        req = self.build_run_deal_req("CS", dealMap, poolAssump, runAssump)

        if debug:
            return req
        
        result = self._send_req(req, url, timeout=30)

        if result is None or 'error' in result or "Left" in set(tz.concat([ _.keys() for _ in result.values()])):
            leftVal = { k:v['Left'] for k,v in result.items() if "Left" in v }
            raise AbsboxError(f"❌ Failed to get response from run: {leftVal}")

        result = tz.valmap(lambda x:x['Right'] ,result)

        assert isinstance(result, dict), f"Result should be a dict but got {type(result)}, {result}"

        rawWarnMsgByScen = {"^".join(k): self._getWarningMsg(v[RunResp.LogResp.value],showWarning) for k, v in result.items()}
        for scen, msgs in rawWarnMsgByScen.items():
            if len(msgs)>0:
                console.print(f"Warning Message from server for {scen}:"+"\n".join(msgs))

        if read:
            return {
                tuple(k.split("^")): dealMap[k.split("^")[0]].read(v) for k,v in result.items()
            }
        else:
            return result

    def runFirstLoss(self, deal, bName, poolAssump=None, runAssump=[], read=True, debug=False) -> dict:
        """run first loss with deal and pool assumptions

        :param deal: a deal object
        :type deal: Generic | SPV
        :param poolAssump: pool performance assumption, a tuple for single run/ a dict for multi-scenario run, defaults to None
        :type poolAssump: tuple, optional
        :param runAssump: deal level assumption, defaults to []
        :type runAssump: list, optional
        :param read: flag to convert result to pandas dataframe, defaults to True
        :type read: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        :return: result of run, a dict of dataframe if `read` is True.
        :rtype: dict
        """
        return self.runRootFinder(deal, poolAssump, runAssump, ("firstLoss", bName), read, debug)

    def runRootFinder(self, deal, poolAssump, runAssump, p, read=True, debug=False) -> dict:
        """run root finder with deal and pool assumptions
        :param deal: a deal object
        :type deal: Generic | SPV
        :param poolAssump: pool performance assumption, a tuple for single run/ a dict for multi-scenario run, defaults to None
        :type poolAssump: tuple, optional
        :param runAssump: deal level assumption, defaults to []
        :type runAssump: list, optional
        :param p: a tuple of root finder parameters
        :type p: tuple/string
        :param read: flag to convert result to pandas dataframe, defaults to True
        :type read: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        :return: result of run
        :rtype: dict
        """

        url = f"{self.url}/{Endpoints.RunRootFinder.value}"
        req = None
        match p:
            case ("firstLoss", bn):
                req = self.build_run_deal_req(("RootFinder", "stressDefault", ("bondIncurLoss", bn)), deal, poolAssump, runAssump)
            case ("maxSpreadToFace", bn, bFlag, fFlag):
                req = self.build_run_deal_req(("RootFinder",("maxSpread", bn), ("bondPricingEqOriginBal", bn, bFlag, fFlag)), deal, poolAssump, runAssump)
            case (tweak,stop):
                req = self.build_run_deal_req(("RootFinder", tweak, stop), deal, poolAssump, runAssump)
        if debug:
            return req

        result = self._send_req(req, url)

        if result is None or 'error' in result or 'Left' in result:
            leftVal = result.get("Left","")
            raise AbsboxError(f"❌ Failed to get response from run: {leftVal}")
        if read:
            return readAeson(result['Right'])
        else:
            return result['Right']



    def runAsset(self, date, _assets, poolAssump=None, rateAssump=None
                 , pricing=None, read=True, debug=False) -> tuple:
        """run asset with assumptions

        :param date: date of start projection and pricing day
        :type date: date
        :param _assets: a list of asset objects
        :type _assets: list
        :param poolAssump: asset performance, defaults to None
        :type poolAssump: tuple, optional
        :param rateAssump: interest rate assumption, defaults to None
        :type rateAssump: tuple, optional
        :param pricing:  princing input for asset, defaults to None
        :type pricing: tuple, optional
        :param read: whether convert raw result to dataframe, defaults to True
        :type read: bool, optional
        :param debug: return request text instead of sending out such request, defaults to False
        :type debug: bool, optional
        :return: (cashflow, balance, pricing result)
        :rtype: tuple
        """
        assert isinstance(_assets, list), f"Assets passed in must be a list"
        
        def readResult(x):
            try:
                (cfs, pr) = x
                cfs = _read_cf(cfs['contents'][1], self.lang)
                pricingResult = _read_asset_pricing(pr, self.lang) if pr else None
                return (cfs, pricingResult)
            except Exception as e:
                print(f"Failed to read result {x} \n with error {e}")
                return (None,None)
            
        url = f"{self.url}/{Endpoints.RunAsset.value}"
        _assumptions = mkAssumpType(poolAssump) if poolAssump else None
        _rate = lmap(mkRateAssumption, rateAssump) if rateAssump else None
        _pricing = mkLiqMethod(pricing) if pricing else None
        assets = lmap(mkAssetUnion, _assets) 
        req = json.dumps([date, assets, _assumptions, _rate, _pricing], ensure_ascii=False)
        if debug:
            return req
        
        result = self._send_req(req, url)

        if result is None or 'error' in result or 'Left' in result:
            leftVal = result.get("Left","")
            raise AbsboxError(f"❌ Failed to get response from run: {leftVal}")
        
        result = result['Right']
        if read:
            return readResult(result)
        else:
            return result

    def runDates(self, d, dp, eDate=None):
        """generate a list of dates from date pattern

        :param d: a starting date
        :type d: date
        :param dp: a date pattern
        :type dp: date pattern
        :return: a list of dates
        :rtype: list[date]
        """
        url = f"{self.url}/{Endpoints.RunDate.value}"
        req = json.dumps([d, mkDatePattern(dp), eDate], ensure_ascii=False)
        result = self._send_req(req, url)
        return result

    def _send_req(self, _req, _url: str, timeout=10, headers={})-> dict | None:
        """generic function send request to server

        :meta private:
        :param _req: request body
        :type _req: string
        :param _url: engine server url
        :type _url: str
        :param timeout: timeout in seconds, defaults to 10
        :type timeout: int, optional
        :param headers: default request header, defaults to {}
        :type headers: dict, optional
        :return: response in dict
        :rtype: dict | None
        """
        assert _req is not None, f"❌request body is None, please check your request"
        try:
            hdrs = self.hdrs | headers
            r = None
            if self.session:
                r = self.session.post(_url, data=_req.encode('utf-8'), headers=hdrs, verify=False, timeout=timeout)
            else:
                raise AbsboxError("❌: None type for session")
        except (ConnectionRefusedError, ConnectionError):
            raise AbsboxError(f"❌ Failed to talk to server {_url}")
        except ReadTimeout:
            raise AbsboxError("❌ Failed to get response from server")
        if r.status_code != 200:
            raise EngineError(r)
        try:
            return json.loads(r.text)
        except JSONDecodeError as e:
            raise EngineError(e)



import logging, os, re, itertools
import requests, shutil, json
from dataclasses import dataclass, field
import functools, pickle, collections
import pandas as pd
import numpy as np
from urllib.request import unquote
from functools import reduce 
import toolz as tz
from .base import *
from .util import *
from .component import *
from .interface import mkTag

def readBondStmt(respBond):
    match respBond:
        case {'tag':'BondGroup','contents':bndMap }:
            return {k: pd.DataFrame(list(tz.pluck("contents",[] if v['bndStmt'] is None else v['bndStmt'])), columns=english_bondflow_fields).set_index("date") for k,v in bndMap.items() }
        case {'tag':'Bond', **singleBndMap }:
            bStmt = mapNone(singleBndMap.get('bndStmt',[]),[])
            return pd.DataFrame(list(tz.pluck("contents", bStmt)), columns=english_bondflow_fields).set_index("date")
        case _:
            raise RuntimeError("Failed to read bond flow from resp",respBond)

def readTrgStmt(x):
    tStmt = mapNone(x.get('trgStmt',[]),[])
    return pd.DataFrame(list(tz.pluck("contents", tStmt)), columns=china_trigger_flow_fields_d).set_index("日期")


@dataclass
class SPV:
    名称: str
    日期: dict
    资产池: dict
    账户: tuple
    债券: tuple
    费用: tuple
    分配规则: dict
    归集规则: tuple
    流动性支持: dict = None
    利率对冲: dict = None
    汇率对冲: dict = None
    触发事件: dict = None
    状态: str = "摊销"
    自定义: dict = None
    科目: dict = None

    @property
    def json(self):
        parsedDates = mkDate(self.日期)
        defaultStartDate = self.日期.get("起息日", None) or self.日期['归集日'][0]
        mixedAssetFlag = isMixedDeal(self.资产池)
        (lastAssetDate,lastCloseDate) = getStartDate(self.日期)
        """
        get the json formatted string
        """
        _r = {
            "dates": parsedDates,
            "name": self.名称,
            "status": mkStatus(self.状态),
            "pool":  mkPoolType(lastAssetDate, self.资产池, mixedAssetFlag),
            "bonds": {bn: mkBndComp(bn, bo) for (bn, bo) in self.债券 },
            "waterfall": mkWaterfall({},self.分配规则.copy()),
            "fees": {fn: mkFee(fo|{"名称":fn}) for (fn, fo) in self.费用 },
            "accounts": {an:mkAcc(an,ao) for (an, ao) in self.账户 },
            "collects": [ mkCollection(c) for c in self.归集规则],
            "rateSwap": {k:mkRateSwap(v) for k,v in self.利率对冲.items()} if self.利率对冲 else None,
            "currencySwap": None,
            "custom": {cn:mkCustom(co) for cn,co in self.自定义.items()} if self.自定义 else None,
            "triggers": renameKs2({k: {_k: mkTrigger(_v) for (_k,_v) in v.items()} for (k,v) in self.触发事件.items() },chinaDealCycle) if self.触发事件 else None,
            "liqProvider": {ln: mkLiqProvider(ln, lo) for ln,lo in self.流动性支持.items() } if self.流动性支持 else None,
            "ledgers": {ln: mkLedger(ln, v) for ln,v in self.科目.items()} if self.科目 else None,
            "stats": [{},{},{},{}]
        }
        _dealType = identify_deal_type(_r)

        return mkTag((_dealType,_r))

    def _get_bond(self, bn):
        for _bn,_bo in self.债券:
            if _bn == bn:
                return _bo
        return None

    def read_pricing(self, pricing):
        if pricing:
            return mkPricingAssump(pricing)
        return None
    
    @staticmethod
    def read(resp):
        read_paths = { #'bonds': ('bndStmt', china_bondflow_fields, "债券")
                     'fees': ('feeStmt', china_fee_flow_fields_d, "费用")
                    , 'accounts': ('accStmt', china_acc_flow_fields_d , "账户")
                    , 'liqProvider': ('liqStmt', china_liq_flow_fields_d, "流动性支持")
                    , 'rateSwap': ('rsStmt', china_rs_flow_fields_d, "")
                    , 'ledgers': ('ledgStmt', china_ledger_flow_fields_d, "")
                    }
        assert isinstance(resp,list),f"<read>:resp should be list,but it is {type(resp)} => {resp}"
        deal_content = resp[0]['contents']
        output = {}
        for comp_name, comp_v in read_paths.items():
            if (not comp_name in deal_content) or (deal_content[comp_name] is None):
                continue
            output[comp_name] = {}
            for k, x in deal_content[comp_name].items():
                ir = None
                if x[comp_v[0]]:
                    ir = list(tz.pluck('contents', x[comp_v[0]]))
                    output[comp_name][k] = pd.DataFrame(ir, columns=comp_v[1]).set_index("日期")
            output[comp_name] = collections.OrderedDict(sorted(output[comp_name].items()))
        # aggregate fees
        output['fees'] = {f: v.groupby('日期').agg({"余额": "min", "支付": "sum", "剩余支付": "min"})
                          for f, v in output['fees'].items()}

        
        # read bonds
        output['bonds'] = {k :readBondStmt(v) for k,v in deal_content['bonds'].items()}

        # trigger 
        if 'triggers' in deal_content and deal_content['triggers']:
            output['triggers'] = deal_content['triggers'] & lens.Values().Values().modify(readTrgStmt)    
        else:
            output['triggers'] = None
        
        # aggregate accounts
        output['agg_accounts'] = aggAccs(output['accounts'], 'chinese')

        output['pool'] = {}
        poolMap = deal_content['pool']['contents']
        
        if deal_content['pool']['tag']=='MultiPool':
            output['pool']['flow'] = tz.valmap(lambda v: readPoolCf(v['futureCf'][0]['contents']) if (not v['futureCf'] is None) else pd.DataFrame(), poolMap)
            output['pool']['breakdown'] = tz.valmap(lambda v: list(tz.map(readPoolCf, v['futureCf'][1] & lens.Each()['contents'].collect() )) if (v['futureCf'] and (not v['futureCf'][1] is None)) else [], poolMap)
        elif deal_content['pool']['tag']=='ResecDeal':
            output['pool']['flow'] = {tz.get([1,2,4],k.split(":")): readPoolCf(v['futureCf']['contents']) for (k,v) in poolMap.items() }
        else:
            raise RuntimeError(f"Failed to match deal pool type:{deal_content['pool']['tag']}")

        outstanding_pool_flow = {k:{"flow": readPoolCf(aggFlow['contents'])
                                    ,"breakdown": [ readPoolCf(_['contents']) for _ in breakdownFlows]}
                                   for k,(aggFlow,breakdownFlows) in resp[4].items()}
        output['pool_outstanding'] = {"flow": { k:v['flow'] for k,v in outstanding_pool_flow.items() }
                                      ,"breakdown": { k:v['breakdown'] for k,v in outstanding_pool_flow.items() } }
 

        output['pricing'] = readPricingResult(resp[3], 'cn')
        output['result'] = readRunSummary(resp[2], 'cn')
        output['_deal'] = resp[0]
        return output



信贷ABS = SPV # Legacy ,to be deleted



from dataclasses import dataclass
from datetime import datetime
import pickle
from rich.console import Console
from rich import print_json
import enum
import getpass ,json
import pandas as pd
from .exception import *
import requests
import toolz as tz

from requests.exceptions import ReadTimeout
from json import JSONDecodeError

from .client import VERSION_NUM
from .local.generic import Generic

console = Console()

from .validation import vStr

__all__ = ["LibraryEndpoints",]

class LibraryEndpoints(str, enum.Enum):
    """Endpoints for deal library"""
    Ack = "ack"
    Token = "token"
    Query = "query"
    Run = "run" # run a deal from the library
    Add = "add" # add new deal to library
    Cmd = "cmd" # add new deal to library
    Get = "get" # get deal from library


class LibraryPath(str, enum.Enum):
    """ Enum class representing shortcut to deal library and data service """
    LDN = "https://ldn.spv.run/api"


@dataclass
class LIBRARY:
    """ Deal Library class for deal library operations 
    :return: Deal Library instance
    :rtype: DealLibrary
    """
    url: str = ""
    token = None
    hdrs = {'Content-type': 'application/json', 'Accept': '*/*', 'Accept-Encoding': 'gzip'}
    debug = False
    libraryInfo = None
    session = None

    def __post_init__(self):

        try:
            _r = requests.post(f"{self.url}/ack", verify=False)
        except Exception as e:
            raise AbsboxError(f"❌ Failed to connect to library server:{e}")
        if _r.status_code == 200:
            self.session = requests.Session()
            self.libraryInfo = json.loads(_r.text)
            console.print(f"✅ Connected to library server")
            console.print(f"absbox version:{self.libraryInfo['absbox']}")
            console.print(f"Hastructure:{self.libraryInfo['Hastructure']}")
        else:
            console.print(f"❌ Failed to connect to library server")

    def login(self, user, pw, **q):
        """login to deal library with user and password

            update token attribute if login success
            
        :param user: username
        :type user: string
        :param pw: password
        :type pw: string
        :param q: query parameters:
                    - {"deal_library": <deal library url> }, specify the deal library server;
        :return: None 
        :rtype: None
        """
        deal_library_url = self.url+f"/{LibraryEndpoints.Token.value}"
        cred = {"user": vStr(user), "password": pw}
        r:dict = self._send_req(json.dumps(cred), deal_library_url)
        if 'token' in r:
            console.print(f"✅ Login successfully, user -> {r['user']},group -> {r['group']}")
            self.token = r['token']
        else:
            console.print(f"❌ Failed to login,{r['msg']}")
            self.token = None
            return None
    
    def ack(self):
        """acknowledge library connection

        :return: library information
        :rtype: dict
        """
        return self._send_req("", f"{self.url}/ack")

    def safeLogin(self, user, **q):
        """safe login with user and password in interactive console

        a interactive input is pending after calling this function

        :param user: username
        :type user: string
        """
        try:
            pw = getpass.getpass(prompt=f"Enter your deal library password for {user}:")
            self.login(user, pw, **q)
        except Exception as e:
            raise AbsboxError(f"❌ Failed during library login {e}")

    def query(self, k = {}, read=True):
        """query deal library with bond ids

        :param ks: bond Ids
        :type ks: list
        :param q: query parameters:

                    - if {"read":True}, return a dataframe, else return raw result;
        :return: a list of bonds found in library
        :rtype: pd.DataFrame or raw result
        
        """
        if not hasattr(self, "token"):
            raise AbsboxError(f"❌ No token found , please call loginLibrary() to login")

        deal_library_url = self.url+f"/{LibraryEndpoints.Query.value}"
        result = self._send_req(json.dumps({"q":k})
                                , deal_library_url
                                , headers={"Authorization": f"Bearer {self.token}"})
        console.print(f"✅ Query success: {k}")
        if read:
            if 'data' in result:
                return pd.DataFrame(result['data'], columns=result['header'])
            elif 'error' in result:
                return pd.DataFrame([result["error"]], columns=["error"])
        else:
            return result


    def get(self, q):
        if not hasattr(self, "token"):
            raise AbsboxError(f"❌ No token found, please call login() to login")
        deal_library_url = self.url+f"/{LibraryEndpoints.Get.value}"

        r = self._send_req(pickle.dumps({"q":q}), deal_library_url
                            , headers={"Authorization": f"Bearer {self.token}"
                                        ,"Content-Type":"application/octet-stream"})
        
        return r
        #console.print(f"✅ Get deal success")

    def run(self, _id, **p):
        """send deal id with assumptions to remote server and get result back

        :param _id: how to find a deal in library database
        :type _id: dict
        :param p: run parameters:
                    - {"read":True}, return a dataframe, else return raw result;
                    - {"poolAssump":<pool assumptions>}, specify pool assumptions;
                    - {"runAssump":<deal assumptions>}, specify run assumptions;
        :type p: dict
        :raises RuntimeError: _description_
        :return: raw string or dataframe
        :rtype: string | pd.DataFrame
        """
        if not hasattr(self, "token"):
            raise AbsboxError(f"❌ No token found, please call login() to login")

        deal_library_url = self.url+f"/{LibraryEndpoints.Run.value}"
        read = p.get("read", True)
        
        runAssump, poolAssump = tz.get(["runAssump", "poolAssump"], p, None)
        runType = p.get("runType", "S")

        runReq = {"q": _id, "runType": runType 
                ,"runAssump": runAssump, "poolAssump": poolAssump
                ,"engine": p.get("engine",None) 
                ,"tweak": p.get("tweak",None),"stop": p.get("stop",None)
                }
        
        bRunReq = pickle.dumps(runReq)

        r = self._send_req(bRunReq, deal_library_url
                            , headers={"Authorization": f"Bearer {self.token}"
                                        ,"Content-Type":"application/octet-stream"})
        if 'error' in r:
            raise AbsboxError(f"❌ error from server:{r['error']}")
        try:
            result = r['result']
            runInfo = tz.dissoc(r, 'result')
            console.print(f"✅ Run success with deal: {runInfo['deal']['name']}|{runInfo['deal']['period']}|{runInfo['deal']['id']}")
        except Exception as e:
            raise AbsboxError(f"❌ message from API server:\n,{e}")
        try:       
            if read and isinstance(result, list):
                return (runInfo, Generic.read(result))
            elif read and isinstance(result, dict):
                return (runInfo, tz.valmap(Generic.read, result))
            else:
                return (runInfo, result)
        except Exception as e:
            raise AbsboxError(f"❌ Failed to read result with error = {e}")


    def _send_req(self, _req, _url: str, timeout=10, headers={})-> dict | None:
        """generic function send request to server

        :meta private:
        :param _req: request body
        :type _req: string
        :param _url: engine server url
        :type _url: str
        :param timeout: timeout in seconds, defaults to 10
        :type timeout: int, optional
        :param headers: default request header, defaults to {}
        :type headers: dict, optional
        :return: response in dict
        :rtype: dict | None
        """
        try:
            hdrs = self.hdrs | headers
            r = None
            if self.session and not isinstance(_req, bytes):
                r = self.session.post(_url, data=_req.encode('utf-8'), headers=hdrs, verify=False, timeout=timeout)
            elif self.session :
                r = self.session.post(_url, data=_req, headers=hdrs, verify=False, timeout=timeout)
            else:
                raise AbsboxError(f"❌ None type for session")
        except (ConnectionRefusedError, ConnectionError):
            raise AbsboxError(f"❌ Failed to talk to server {_url}")
        except ReadTimeout:
            raise AbsboxError(f"❌ Failed to get response from server")
        if r.status_code != 200:
            raise EngineError(r)
        try:
            return json.loads(r.text)
        except JSONDecodeError as e:
            raise EngineError(e)



from schema import Schema,Regex,Or

from urllib.parse import urlparse


def isValidUrl(url: str) -> str | None:
    try:
        result = urlparse(url)
        if all([result.scheme, result.netloc]):
            return url
    except ValueError:
        print(f"Invalid URL {url}")
        return None


dateStr = Regex(r"^\d{4}-\d{2}-\d{2}$")


def vList(x, t, msg:str = None) -> list:
    return Schema([t]).validate(x)

def vListOfList(x, t, msg:str = None) -> list:
    return Schema([[t]]).validate(x)

def vTuple(x, t, msg:str = None) -> tuple:
    return Schema((t,)).validate(x)

def vDict(x, msg:str = None) -> dict:
    pass


def vStr(x, msg:str = None) -> str:
    return Schema(str).validate(x)


def vNum(x, msg:str = None) -> float:
    return Schema(Or(float, int)).validate(x)

def vFloat(x, msg:str = None) -> float:
    return Schema(float).validate(x)


def vInt(x, msg:str = None) -> int:
    return Schema(int).validate(x)


def vBool(x, msg:str = None) -> bool:
    return Schema(bool).validate(x)


def vDate(x, msg:str = None) -> str:
    return Schema(dateStr).validate(x)


def vCurve(x, msg:str = None):
    return Schema([[Schema(dateStr), Or(int, float)]]).validate(x)


def vTable(x, msg:str = None):
    return Schema([[Or(int, float) ,Or(int, float)]]).validate(x)


def validation(deal):
    errors = []
    warnings = []
    if len(errors) > 0:
        return False, errors, warnings
    else:
        return True, [], warnings


from dataclasses import dataclass
import functools
import pandas as pd
import collections
import toolz as tz

from .util import mapListValBy,mapValsBy,renameKs2\
                        ,guess_pool_flow_header,positionFlow,mapNone
from .util import earlyReturnNone,lmap,isMixedDeal                              
from .component import *
from .base import * 
from ..validation import vStr,vDate,vNum,vList,vBool,vFloat,vInt
from .interface import mkTag,readAeson

def readBondStmt(respBond):
    match respBond:
        case {'tag':'BondGroup','contents': [bndMap, pt] }:
            return {k: pd.DataFrame(list(tz.pluck("contents",[] if v['bndStmt'] is None else v['bndStmt'])), columns=english_bondflow_fields).set_index("date") for k,v in bndMap.items() }
        case {'tag':'Bond', **singleBndMap }:
            bStmt = mapNone(singleBndMap.get('bndStmt',[]),[])
            return pd.DataFrame(list(tz.pluck("contents", bStmt)), columns=english_bondflow_fields).set_index("date")
        case {'tag':'MultiIntBond', **singleBndMap }:
            bStmt = mapNone(singleBndMap.get('bndStmt',[]),[])
            return pd.DataFrame(list(tz.pluck("contents", bStmt)), columns=english_bondflow_fields).set_index("date")
        case _:
            raise RuntimeError("Failed to read bond flow from resp",respBond)

def readTrgStmt(x):
    tStmt = mapNone(x.get('trgStmt',[]),[])
    return pd.DataFrame(list(tz.pluck("contents", tStmt)), columns=english_trigger_flow_fields_d).set_index("date")

@dataclass
class Generic:
    name: str
    dates: dict
    pool: dict
    accounts: tuple
    bonds: tuple
    fees: tuple
    waterfall: dict
    collection: list
    liqFacility: dict = None
    rateSwap: dict = None
    currencySwap: dict = None
    trigger: dict = None
    status: str = "Amortizing"
    custom: dict = None
    ledgers: dict = None
    rateCap: dict = None
    stats: dict = None

    @property
    def json(self) -> dict:
        parsedDates = mkDate(self.dates)
        (lastAssetDate, lastCloseDate) = getStartDate(self.dates)
        mixedAssetFlag = isMixedDeal(self.pool)
        """
        get the json formatted string
        """
        _r = {
            "dates": parsedDates,
            "name": vStr(self.name),
            "status":mkStatus(self.status),
            "pool":mkPoolType(lastAssetDate, self.pool, mixedAssetFlag),
            "bonds": {bn: mkBndComp(bn, bo) for (bn, bo) in self.bonds},
            "waterfall": mkWaterfall({},self.waterfall.copy()),  
            "fees": {fn: mkFee(fo|{"name": fn}) for (fn, fo) in self.fees},
            "accounts": {an:mkAcc(an, ao) for (an, ao) in self.accounts},
            "collects": lmap(mkCollection, self.collection),
            "rateSwap": tz.valmap(mkRateSwap, self.rateSwap) if self.rateSwap else None,
            "rateCap": tz.valmap(mkRateCap, self.rateCap) if self.rateCap else None,
            "currencySwap":None ,
            "custom": tz.valmap(mkCustom, self.custom) if self.custom else None,
            "triggers": renameKs2({k: {_k: mkTrigger(_v) for (_k,_v) in v.items() } for (k, v) in self.trigger.items()},englishDealCycle) if self.trigger else None,
            "liqProvider": {ln: mkLiqProvider(ln, lo) 
                            for ln,lo in self.liqFacility.items() } if self.liqFacility else None,
            "ledgers": {ln: mkLedger(ln, v) for ln,v in self.ledgers.items()} if self.ledgers else None,
            "stats": self.buildStats()
        }

        _dealType = identify_deal_type(_r)

        return mkTag((_dealType, _r))

    def buildStats(self,):
        if self.stats is None:
            return [{},{},{},{}]
        a = {k:v for k,v in self.stats.items() if k in dealStatBalance }
        b = {k:v for k,v in self.stats.items() if k in dealStatRate }
        c = {k:v for k,v in self.stats.items() if k in dealStatBool }
        d = {k:v for k,v in self.stats.items() if k in dealStatInt }
        
        # bal, rate, bool, int
        return [a,b,c,d]

    def read_pricing(self, pricing):
        return earlyReturnNone(mkPricingAssump, pricing)
    
    @staticmethod
    def read(resp):
        read_paths = { 'fees': ('feeStmt', english_fee_flow_fields_d, "fee")
                     , 'accounts': ('accStmt', english_acc_flow_fields_d, "account")
                     , 'liqProvider': ('liqStmt', english_liq_flow_fields_d, "")
                     , 'rateSwap': ('rsStmt', english_rs_flow_fields_d, "")
                     , 'rateCap': ('rcStmt', english_rs_flow_fields_d, "")
                     , 'ledgers': ('ledgStmt', english_ledger_flow_fields_d, "")
                     }
        output = {}
        output['_deal'] = readAeson(resp[0])
        deal_content = resp[0]['contents']

        for comp_name, comp_v in read_paths.items():
            if deal_content[comp_name] is None:
                continue
            output[comp_name] = {}
            for k, x in deal_content[comp_name].items():
                ir = None
                if x[comp_v[0]]:
                    ir = list(tz.pluck('contents', x[comp_v[0]]))
                    output[comp_name][k] = pd.DataFrame(ir, columns=comp_v[1]).set_index("date")
                elif x[comp_v[0]] is None:
                    output[comp_name][k] = pd.DataFrame([], columns=comp_v[1]).set_index("date")
                else:
                    pass
            output[comp_name] = collections.OrderedDict(sorted(output[comp_name].items()))
        # aggregate fees
        output['fees'] = {f: v.groupby('date').agg({"balance": "min", "payment": "sum", "due": "min"})
                          for f, v in output['fees'].items()}

        # read bonds
        output['bonds'] = {k :readBondStmt(v) for k,v in deal_content['bonds'].items()}

        # triggers
        if 'triggers' in deal_content and deal_content['triggers']:
            output['triggers'] = deal_content['triggers'] & lens.Values().Values().modify(readTrgStmt)    
        else:
            output['triggers'] = None

        # aggregate accounts
        output['agg_accounts'] = aggAccs(output['accounts'], 'english')
        
        output['pool'] = {}
        outstanding_pool_flow = {k:{"flow": readPoolCf(aggFlow['contents'])
                                    ,"breakdown": [ readPoolCf(_['contents']) for _ in breakdownFlows ] if breakdownFlows else []}
                                   for k,(aggFlow,breakdownFlows) in resp[4].items()}
        output['pool_outstanding'] = {"flow": { k:v['flow'] for k,v in outstanding_pool_flow.items() }
                                      ,"breakdown": { k:v['breakdown'] for k,v in outstanding_pool_flow.items() } }
        
        poolMap = deal_content['pool']['contents']
        
        if deal_content['pool']['tag']=='MultiPool':
            output['pool']['flow'] = tz.valmap(lambda v: readPoolCf(v['futureCf'][0]['contents']) if (not v['futureCf'] is None) else pd.DataFrame(), poolMap)
            output['pool']['breakdown'] = tz.valmap(lambda v: list(tz.map(readPoolCf, v['futureCf'][1] & lens.Each()['contents'].collect() )) if (v['futureCf'] and (not v['futureCf'][1] is None)) else [], poolMap)
        elif deal_content['pool']['tag']=='ResecDeal':
            output['pool']['flow'] = {tz.get([1,2,4],k.split(":")): readPoolCf(v['futureCf']['contents']) for (k,v) in poolMap.items() }
        else:
            raise RuntimeError(f"Failed to match deal pool type:{deal_content['pool']['tag']}")

        output['pricing'] = readPricingResult(resp[3], 'en')
        output['result'] = readRunSummary(resp[2], 'en')

        return output

    def __str__(self):
        ''' Return a simplified string representation of deal object '''
        return f"{self.name}"



from absbox import mkDeal

deal_data = {
    "name":"Multiple Pools with Mixed Asset"
    ,"dates":{"cutoff":"2021-06-01"
              ,"closing":"2021-07-15"
              ,"firstPay":"2021-08-26"
              ,"payFreq":["DayOfMonth",20]
              ,"poolFreq":"MonthEnd"
              ,"stated":"2030-01-01"}
    ,"pool":{"PoolA":{'assets':[["Mortgage"
                        ,{"originBalance":2200,"originRate":["fix",0.045],"originTerm":30
                          ,"freq":"Monthly","type":"Level","originDate":"2021-02-01"}
                          ,{"currentBalance":2200
                          ,"currentRate":0.08
                          ,"remainTerm":30
                          ,"status":"current"}]]},
             "PoolB":{'assets':[["Loan"
                              ,{"originBalance": 80000
                                ,"originRate": ["floater",0.045,{"index":"SOFR3M"
                                                                ,"spread":0.01
                                                                ,"reset":"QuarterEnd"}]
                                ,"originTerm": 60
                                ,"freq": "Monthly"
                                ,"type": "i_p"
                                ,"originDate": "2021-02-01"}
                              ,{"currentBalance": 65000
                                ,"currentRate": 0.06
                                ,"remainTerm": 60
                                ,"status": "Current"}]]}
            }
    ,"accounts":{"acc01":{"balance":0}
                ,"acc02":{"balance":0}}
    ,"bonds":{"A1":{"balance":1000
                 ,"rate":0.07
                 ,"originBalance":1000
                 ,"originRate":0.07
                 ,"startDate":"2020-01-03"
                 ,"rateType":{"Fixed":0.08}
                 ,"bondType":{"Sequential":None}}
             ,"B":{"balance":1000
                     ,"rate":0.0
                     ,"originBalance":1000
                     ,"originRate":0.07
                     ,"startDate":"2020-01-03"
                     ,"rateType":{"Fixed":0.00}
                     ,"bondType":{"Equity":None}}}
    ,"fees":{"trusteeFee":{"type":{"fixFee":30},"feeStart":"2021-07-15"}
            ,"serviceFee":{"type":{"annualPctFee":[("poolBalance","PoolB"),0.02]},"feeStart":"2021-07-15"}
            ,"serviceFee2":{"type":{"byTable":["MonthEnd"
                                               ,("const",1)
                                               ,[(0,5),(2,10),(10,15)]]
                                   },"feeStart":"2021-07-15"}}
    ,"collect":[[["PoolA"],"CollectedCash","acc01"]
               ,[["PoolB"],"CollectedCash","acc02"]]
    ,"waterfall":{"Amortizing":[
         ["payFee","acc01",['trusteeFee']]
         ,["calcAndPayFee","acc01",['serviceFee']]
         ,["accrueAndPayInt","acc01",["A1"]]
         ,["payPrin","acc01",["A1"]]
         ,["payPrin","acc01",["B"]]
         ,["payPrinResidual","acc01",["B"]]
     ]}
    ,"status":("PreClosing","Amortizing")
}


mixedAsset_test01 = mkDeal(deal_data)




class VersionMismatch(Exception):
    """Exception for version mismatch between client and server"""
    def __init__(self, libVersion, serverVersion) -> None:
        self.libVersion = libVersion
        self.serverVersion = serverVersion
        super().__init__(f"Failed to match version, lib support={libVersion} but server version={serverVersion}")

class EngineError(Exception):
    """Exception for error from engine server"""
    def __init__(self, engineResp) -> None:
        errorMsg = engineResp.text
        super().__init__(errorMsg)

class AbsboxError(Exception):
    """Exception for error from absbox"""
    def __init__(self, errorMsg) -> None:
        super().__init__(errorMsg)

class LibraryError(Exception):
    """Exception for error from absbox"""
    def __init__(self, errorMsg) -> None:
        super().__init__(errorMsg)


import pandas as pd 
from .base import china_bondflow_fields_s, english_bondflow_fields_s,english_bondflow_cash,china_bondflow_cash
from ..validation import vStr
from toolz import get_in

def runYieldTable(api, d, bondName:str, p_assumps: dict, b_assumps: dict):
    """ a shortcut function to run a 'yield table' for a bond, with different pool assumptions

    :param api: the api object
    :type api: _type_
    :param d: the deal object
    :type d: _type_
    :param bondName: the bond name
    :type bondName: str
    :param p_assumps: a map of pool assumptions
    :type p_assumps: dict
    :param b_assumps: pricing assumption for bond
    :type b_assumps: tuple
    :return: _description_
    :rtype: _type_
    """
    assert isinstance(p_assumps, dict), f"pool assumption should be a map but got {type(p_assumps)}"
    
    rs = api.runByScenarios(d, poolAssump=p_assumps
                            , runAssump=[('pricing', b_assumps)]
                            , read=True)

    b_pricing = {k: v['pricing'].loc[ vStr(bondName)] for k, v in rs.items()}
    b_table = pd.concat(b_pricing.values(), axis=1)
    b_table.columns = list(b_pricing.keys())

    return b_table

run_yield_table = runYieldTable


def flow_by_scenario(rs, flowpath, node="col", rtn_df=True, ax=1, rnd=2):
    "pull flows from multiple scenario"
    r = None
    if node == "col":
        r = {k: get_in(flowpath[:-1], v)[flowpath[-1]] for k, v in rs.items()}    
    elif node == "idx":
        r = {k: get_in(flowpath[:-1], v).loc[flowpath[-1]] for k, v in rs.items()}    
    else:
        r = {k: get_in(flowpath, v) for k, v in rs.items()}
    if rtn_df:
        _vs = list(r.values())
        _ks = list(r.keys())
        r = pd.concat(_vs, keys=_ks, axis=ax)
    return r


import pandas as pd
import functools,json,copy,logging,re,itertools
from functools import reduce
from datetime import datetime
from lenses import lens, ui, optics
import toolz as tz

from .base import *
from .interface import mkTag


def mapNone(x, v):
    ''' return a default value (v) if x is None, other wise, return x '''
    if x is None:
        return v
    else:
        return x


def lmap(f: callable, xs) -> list:
    ''' just make it looks more functional '''
    return list(map(f, xs))


def flat(xss) -> list:
    return reduce(lambda xs, ys: xs + ys, xss)


def filter_by_tags(xs: list, tags: list) -> list:
    ''' fiter a list of maps by tags'''
    tags_set = set(tags)
    return [x for x in xs if x['tag'] in tags_set]


def readTagStr(x: str) -> str:
    x = json.loads(x.replace("'", "\"").replace("True", "true").replace("False", "false").replace("None","null"))
    return readTagMap(x)


def readTagMap(x:dict) -> str:
    match x:
        case {"tag": _t, "contents": None}:
            return f"<{_t}>"
        case {"tag": _t, "contents": _c} if isinstance(_c, list):
            return f"<{_t}:{','.join([ readTagMap(_) for _ in _c])}>"
        case {"tag": _t,"contents":_c} if isinstance(_c, str):
            return f"<{_t}:{_c}>"
        case {"tag": _t,"contents":_c} if isinstance(_c, dict):
            return f"<{_t}:{readTagMap(_c)}>"
        case {"tag": _t}:
            return f"<{_t}>"
        case _ :
            return f"<{x}>"


def readTag(x: dict):
    return f"<{x['tag']}:{','.join(x['contents'])}>"


def isDate(x):
    return re.match(r"\d{4}\-\d{2}\-\d{2}", x)


def allList(xs):
    return all([isinstance(x, list) for x in xs])


def mkTs(n, vs):
    return mkTag((n, vs))


def mkTbl(n, vs):
    return mkTag((n, vs))


mkRatioTs = functools.partial(mkTs, "RatioCurve")
mkRateTs = functools.partial(mkTs, "IRateCurve")
mkBalTs = functools.partial(mkTs, "BalanceCurve")
mkFloatTs = functools.partial(mkTs, "FloatCurve")


def unify(xs, ns):
    "union dataframes by stacking up with names provided"
    index_name = xs[0].index.name
    dfs = []
    for x, n in zip(xs, ns):
        dfs.append(pd.concat([x], keys=[n], axis=1))
    r = functools.reduce(lambda acc, x: acc.merge(x
                                                , how='outer'
                                                , on=[index_name])
                        , dfs)
    return r.sort_index()


def unifyTs(xs):
    "unify time-series alike dataframe"
    _index_set = set([x.index.name for x in xs])
    assert len(_index_set)==1,f"Index of dataframes should have same name,got:{_index_set}"

    _index_name = list(_index_set)[0]
    r = functools.reduce(lambda acc, x: acc.merge(x
                                                , how='outer'
                                                , on=[_index_name]), xs)
    
    return r.sort_index()


def update_deal(d, i, c): #Deprecated ,to be replace with Lens
    "A patch function to update a deal data component list in immuntable way"
    _d = d.copy()
    _d.pop(i)
    _d.insert(i, c)
    return _d


def str2date(x: str):
    return datetime.strptime(x, '%Y-%m-%d').date()


def normDate(x: str):
    if len(x) == 8:
        return f"{x[:4]}-{x[4:6]}-{x[6:8]}"


def guess_locale(x):
    ''' Guess local from deal map'''
    accs = x['accounts']

    assert len(accs) > 0, "Failed to identify via deal accounts result"

    acc_cols = set(list(accs.values())[0].columns.to_list())
    locale = None
    if acc_cols == set(["余额", "变动额", "备注"]):
        locale = "cn"
    if acc_cols == set(["balance", "change", "memo"]):
        locale = "en"
    return locale


def guess_pool_locale(x):
    if "cutoffDate" in x:
        return "english"
    elif '封包日' in x:
        return "chinese"
    else:
        raise RuntimeError("Failed to match {x} in guess pool locale")


def renameKs(_m: dict, mapping, opt_key=False):
    '''
    rename keys in a map with from a mapping tuple passed in 
    `opt_key` = True, allow skipping mapping tuple not exist in the map
    '''
    m = _m.copy()
    for (o, n) in mapping:
        if opt_key and o not in m:
            continue
        m[n] = m[o]
        del m[o]
    return m


def ensureKeysInMap(m: dict, ks: list, msg=""):
    ''' Ensure keys in a map, if not found, add with None'''
    missing = []
    for k in ks:
        if k not in m:
            missing.append(k)
    if missing:
        raise RuntimeError(f"Missing keys in map:{missing}, not found in {m.keys()}, location:{msg}")
    return m



def subMap(m: dict, ks: list):
    ''' get a map subset by keys,if keys not found, supplied with default value '''
    return {k: m.get(k, defaultVal) for (k, defaultVal) in ks}


def subMap2(m: dict, ks: list):
    _m = {k: m.get(k, defaultVal) for (k, _, defaultVal) in ks}
    _mapping = [_[:2] for _ in ks]
    return renameKs(_m, _mapping)


def mapValsBy(m: dict, f: callable):
    ''' Given a map and apply function to every vals'''
    assert isinstance(m, dict), f"M is not a map but a {type(m)}, {m}"
    return {k: f(v) for k, v in m.items()}


def mapListValBy(m: dict, f: callable):
    ''' Given a map, whose vals are list,  apply function to every element in each list of each val'''
    assert isinstance(m, dict), "M is not a map"
    return {k: [f(_v) for _v in v] for k,v in m.items()}


def applyFnToKey(m: dict, f, k, applyNone=False):
    assert isinstance(m, dict), f"{m} is not a map"
    assert k in m, f"{k} is not in map {m}"
    assert callable(f), "funciton passed in is not callable"
    match (m[k], applyNone):
        case (None, True):
            m[k] = f(m[k])
        case (None, False):
            pass
        case (_, _):
            m[k] = f(m[k])
    return m


def renameKs2(m: dict, kmapping:dict) -> dict:
    ''' Given a map, rename ks from a key-mapping '''
    assert isinstance(m, dict), "M is not a map"
    assert isinstance(kmapping, dict), f"Mapping is not a map: {kmapping}"
    assert set(m.keys()).issubset(set(kmapping.keys())), f"{m.keys()} not in {kmapping.keys()}"
    return {kmapping[k]: v for k, v in m.items()}


def updateKs(m: dict, kmapping:dict) -> dict:
    ''' Given a map, update ks from a key-mapping '''
    assert isinstance(m, dict), "M is not a map"
    assert isinstance(kmapping, dict), f"Mapping is not a map: {kmapping}"

    return {kmapping.get(k,k):v for k,v in m.items()}


def ensure100(xs, msg=""):
    assert sum(xs) == 1.0, f"Doesn't not sum up 100%: {msg}"


def guess_pool_flow_header(x, l):
    assert isinstance(x, dict), f"x is not a map but {x}, type:{type(x)}"
    match (x['tag'], len(x['contents']), l):
        case ('MortgageDelinqFlow', 12, 'chinese'):
            return (china_mortgage_delinq_flow_fields_d, "日期", False)
        case ('MortgageDelinqFlow', 13, 'chinese'):
            return (china_mortgage_delinq_flow_fields_d+china_cumStats, "日期", True)
        case ('MortgageDelinqFlow', 12, 'english'):
            return (english_mortgage_delinq_flow_fields_d, "Date", False)
        case ('MortgageDelinqFlow', 13, 'english'):
            return (english_mortgage_delinq_flow_fields_d+english_cumStats, "Date", True)
        case ('MortgageFlow', 11, 'chinese'):
            return (china_mortgage_flow_fields_d, "日期", False)
        case ('MortgageFlow', 12, 'chinese'):
            return (china_mortgage_flow_fields_d+china_cumStats, "日期", True)
        case ('MortgageFlow', 11, 'english'):
            return (english_mortgage_flow_fields_d, "Date", False)
        case ('MortgageFlow', 12, 'english'):
            return (english_mortgage_flow_fields_d+english_cumStats, "Date", True)
        case ('LoanFlow', 9, 'chinese'):
            return (china_loan_flow_d, "日期", False)
        case ('LoanFlow', 10, 'chinese'):
            return (china_loan_flow_d+china_cumStats, "日期", True)
        case ('LoanFlow', 9, 'english'):
            return (english_loan_flow_d, "Date", False)
        case ('LoanFlow', 10, 'english'):
            return (english_loan_flow_d+english_cumStats, "Date", True)
        case ('LeaseFlow', 4, 'chinese'):
            return (china_rental_flow_d, "日期", False)
        case ('LeaseFlow', 4, 'english'):
            return (english_rental_flow_d, "Date", False)
        case ('FixedFlow', 6, 'chinese'):
            return (china_fixed_flow_d, "日期", False)
        case ('FixedFlow', 6, 'english'):
            return (english_fixed_flow_d, "Date", False)
        case ('ReceivableFlow', 9, 'chinese'):
            return (china_receivable_flow_d+china_cumStats, "日期", True)
        case ('ReceivableFlow', 9, 'english'):
            return (english_receivable_flow_d+english_cumStats, "Date", True)
        case ('BondFlow', 4, 'chinese'):
            return (china_uBond_flow_d, "日期", False)
        case ('BondFlow', 4, 'english'):
            return (english_uBond_flow_d, "Date", False)
        case _:
            raise RuntimeError(f"Failed to match pool header with {x['tag']},{len(x['contents'])},{l}")


def inferPoolTypeFromAst(x:dict) -> str:
    match x:
        case {"assets":[["Mortgage",*fields],*ast]} | {"assets":[["AdjustRateMortgage",*fields],*ast]} | {'清单': [['按揭贷款', *fields], *ast]}:
            return "MPool"
        case {"assets":[["Loan",*fields],*ast]} | {'清单': [['贷款', *fields], *ast]}:
            return "LPool"
        case {"assets":[["Installment",*fields],*ast]} | {'清单': [['分期', *fields], *ast]}:
            return "IPool"
        case {"assets":[["Lease",*fields],*ast]} | {'清单': [['租赁', *fields], *ast]}:
            return "RPool"
        case {"assets":[["FixedAsset",*fields],*ast]}:
            return "FPool"
        case {"assets":[["Invoice",*fields],*ast]} | {'清单': [['应收账款', *fields], *ast]}:
            return "VPool"
        case {"assets":[["ProjectedCashflow",*fields],*ast]} | {"assets":[["ProjectedByFactor",*fields],*ast]} :
            return "PPool"
        case _:
            raise RuntimeError(f"Failed to find pool type from assets:{x}")
    


def uplift_m_list(l: list):
    """ input a list of dictionary, reduce the maps into a big map """
    return {k: v for m in l for k, v in m.items()}


def getValWithKs(m: dict, ks: list, defaultReturn=None, mapping=None):
    ''' Get first available key/value in m, with optional mapping function to result if found '''
    r = defaultReturn
    if isinstance(m, dict):
        for k in ks:
            if k in m:
                r = m[k]
                break
    else:  # object instead of dict
        for k in ks:
            if hasattr(m, k):
                r = getattr(m, k)
                break
    if (r is not None) and (mapping is not None):
        return mapping(r)
    return r


def _read_cf(x, lang):
    ''' read cashflow from a list , and set index to date'''
    if x == []:
        return []
    assert isinstance(x, list), f"Input is not a list but {type(x)}"
    flow_header, idx, expandFlag = guess_pool_flow_header(x[0], lang)
    result = None
    try:
        if expandFlag:
            result = pd.DataFrame([_['contents'][:-1]+mapNone(_['contents'][-1],[0,0,0,0,0,0]) for _ in x], columns=flow_header)
        else:
            result = pd.DataFrame([_['contents'] for _ in x], columns=flow_header)
    except ValueError as e:
        logging.error(f"{e}")
        logging.error(f"Failed to match header:{flow_header} with {result}")
        return False
    result.set_index(idx, inplace=True)
    result.index.rename(idx, inplace=True)
    result.sort_index(inplace=True)
    return result


def _read_asset_pricing(xs, lang) -> pd.DataFrame:
    return pd.DataFrame(tz.pluck("contents", xs)
            , columns=assetPricingHeader[lang])


def mergeStrWithDict(s: str, m: dict) -> str:
    ''' merge a map to a string'''
    t = json.loads(s)
    return json.dumps(t | m)


def positionFlow(x, m: dict, facePerPaper=100):
    ''' Get a position bond cashflow from a run result '''
    _, _bflow = list(x['bonds'].items())[0]
    bflowHeader = _bflow.columns.to_list()

    def calcBondFlow(_bflow: pd.DataFrame, factor):
        ''' input: (bond cashflow : dataframe, factor: float) '''
        bflow = copy.deepcopy(_bflow)
        bflow[bflowHeader[0]] *= factor
        bflow[bflowHeader[1]] *= factor
        bflow[bflowHeader[2]] *= factor
        bflow[bflowHeader[4]] *= factor
        return bflow
        
    assert isinstance(m, dict), "Position info must be a map/dict"
    
    bOrignBal = {bondName: x['_deal']['contents']['bonds'][bondName]['bndOriginInfo']['originBalance'] for bondName, bondPos in m.items()}
    bPapersPerBond = {bondName: bondOrigBal/facePerPaper for (bondName, bondOrigBal) in bOrignBal.items()}
    bflowFactor = {bondName: (bondPos/bPapersPerBond[bondName]) for (bondName, bondPos) in m.items()}
    return {bn: calcBondFlow(bf, bflowFactor[bn]) for bn, bf in x['bonds'].items() if bn in m}


def tryConvertTupleToDict(xs):
    "build a dictionary from a tuple"
    if isinstance(xs, tuple):
        return {n: v for n, v in xs}
    elif isinstance(xs, dict):
        return xs
    else:
        raise RuntimeError(f"Failed to match <convertTupleToDict> ,either Map or Tuple but got {type(xs)}")


def allKeysAreString(m: dict):
    return all([isinstance(_, str) for _ in m.keys()])


def earlyReturnNone(fn: callable, v):
    "return None if passed in a None, otherwise apply fn and return"
    if v is None:
        return None
    else:
        return fn(v)


def searchByFst(xs:list, v, defaultRtn=None):
    "search by first element in a list, return None if not found"
    for x in xs:
        if x[0] == v:
            return x
    return defaultRtn


def isMixedDeal(x: dict) -> bool:
    ''' check if pool of deals has mixed asset types'''    
    if 'assets' in x or 'cashflow' in x:
        return False
    if '清单' in x or '归集表' in x:
        return False
    if 'deals' in x:
        return False
    assetTags = x & lens.Values().Each()[1][0][0].collect()
    if len(set(assetTags)) > 1:
        return True
    return False


def strFromPath(xs: list) -> str:
    ps = [strFromLens(_)+f":{v}" for (_, v) in xs]
    return "/".join(ps)


def strFromLens(x) -> str: 
    if isinstance(x, ui.UnboundLens) and not hasattr(x._optic, "lenses"):
        if isinstance(x._optic, optics.true_lenses.GetitemLens):
            return str(x._optic.key)
        return x._optic.name
    elif isinstance(x, ui.UnboundLens) and hasattr(x._optic, "lenses"):
        return "-".join([strFromLens(_) for _ in x._optic.lenses])
    elif isinstance(x, optics.true_lenses.GetitemLens):
        return str(x.key)
    elif isinstance(x, optics.traversals.GetZoomAttrTraversal):
        return x.name   
    else:
        return str(x)


def enumVals(e) -> list:
    ''' return a list of enum values '''
    return [_.value for _ in [*e]]


def readCfFromLst(lst:list)-> pd.DataFrame:
    return None


def tupleToDictWithKey(xs,key="name"):
    return dict([ (n,x|{key:n}) for (n,x) in xs ])


def patchDicts(dict1:dict,dict2:dict)-> dict:
    """
    Merge two dictionaries, if a key exists in both dictionaries,use the value from dict2
    if key only exists in either one, use the only one value
    """
    return tz.merge_with(lambda xs: xs[1] if len(xs)==2 else xs[0] ,dict1 ,dict2)


def getNumCols(df:pd.DataFrame)-> list:
    numeric_columns = [col for col in df.columns if pd.to_numeric(df[col], errors='coerce').notna().all()]
    return numeric_columns



import pandas as pd
import toolz as tz
from lenses import lens
import enum
from ..validation import vInt, vDate, vFloat, vBool
from ..validation import vDict, vList, vStr, vNum
from .util import unifyTs,getNumCols


def readToCf(xs, header=None, idx=None, sort_index=False) -> pd.DataFrame:
    ''' input with flow type json, return a dataframe '''
    rows = [_['contents'] for _ in xs]

    if header:
        r = pd.DataFrame(rows, columns=header)
    else:
        r = pd.DataFrame(rows)

    if idx:
        r = r.set_index(idx)

    if sort_index:
        r = r.sort_index()

    return r


# a slice performance gain on this function
def buildDaySeq(_df):
    """  https://stackoverflow.com/questions/23435270/add-a-sequential-counter-column-on-groups-to-a-pandas-dataframe """
    df = _df.reset_index()
    dfGrp = df['date'].ne(df['date'].shift()).cumsum()
    df['daySeq'] = df['date'].groupby(dfGrp).cumcount()
    return df.set_index(["date", "daySeq"])


def buildDaySeq2(_df):
    """  https://stackoverflow.com/questions/23435270/add-a-sequential-counter-column-on-groups-to-a-pandas-dataframe """
    df = _df.reset_index()
    df['daySeq'] = df.groupby(["date"]).cumcount()
    return df.set_index(["date", "daySeq"])


def filterCols(xs, columnsToKeep):
    return [_[columnsToKeep] for _ in xs]


class BondCfHeader(enum.Enum):
    SIMPLE = ['rate','cash','intDue','intOverInt','factor','memo']+["利率","本息合计","应付利息","罚息","本金系数","备注"]
    STANDARD = ['intOverInt','factor','memo']+["罚息","本金系数","备注"]
    FULL = []


def readBondsCf(bMap, popColumns=["factor","memo","本金系数","备注","应付利息","罚息","intDue","intOverInt"]) -> pd.DataFrame:
    def isBondGroup(k, v) -> bool:
        if isinstance(k,str) and isinstance(v,dict):
            return True
        else:
            return False
    def hasBondGroup():
        _r = [ isBondGroup(k, v) for k, v in bMap.items()]
        return any(_r)

    def filterCols(x:dict, columnsToKeep) -> pd.DataFrame:
        return lens.Recur(pd.DataFrame).modify(lambda z: z[columnsToKeep])(x)

    bondNames = list(bMap.keys())
    if not bondNames:
        return pd.DataFrame()

    # colums of bond from resp
    bondColumns = (bMap & lens.Recur(pd.DataFrame).get()).columns.to_list()

    # columns to show for each bond
    columns = list(filter(lambda x: x not in set(popColumns), bondColumns))
    if not hasBondGroup():
        header = pd.MultiIndex.from_product([bondNames,columns])
        yyz = list(filterCols(bMap, columns).values())
        df = pd.concat(yyz,axis=1)
    else:
        bMap = filterCols(bMap, columns)
        indexes = []
        cfFrame = []
        for k,v in bMap.items():
            if isBondGroup(k,v):
                for _k,_v in v.items(): # sub bonds
                    indexes.extend([(k,_k,_) for _ in columns])
                    cfFrame.append(_v)
            else:
                if bMap[k] is not None:
                    indexes.extend([(k,"-",_) for _ in columns])
                    cfFrame.append(v)
                    
        header = pd.MultiIndex.from_tuples(indexes)       
        
        df = pd.concat(cfFrame,axis=1)
    df.columns = header
    return df.sort_index()


def patchMissingIndex(df,idx:set)-> pd.DataFrame:
    curIdx = set(df.index.values.tolist())
    missingIdx = idx - curIdx
    return df.reindex(df.index.values.tolist()+list(missingIdx)).sort_index()

def buildJointCf(m:dict, popColumns=[]) -> pd.DataFrame:
    if len(m)==0:
        return pd.DataFrame()
    fullColumns = list(m.values())[0].columns.to_list()
    columns = list(filter(lambda x: x not in set(popColumns), fullColumns))
    accNames = list(m.keys())
    accVals = [buildDaySeq(av) for av in list(m.values())]

    accIndex = set(tz.concat([ a.index.values.tolist() for a in accVals]))

    accVals2 = [patchMissingIndex(av,accIndex) for av in accVals]
    
    r = pd.concat(accVals2, axis=1, keys=tuple(accNames))
    if popColumns:
        return r.xs(slice(*columns), level=1, axis=1, drop_level=False)
    return r


def readAccsCf(aMap, popColumns=["memo"]) -> pd.DataFrame:
    return buildJointCf(aMap, popColumns=popColumns)


def readFeesCf(fMap, popColumns=["due","剩余支付"]) -> pd.DataFrame:
    return buildJointCf(fMap, popColumns=popColumns)


def readLedgers(lMap, **k) -> pd.DataFrame:
    return buildJointCf(lMap, **k)


def readPoolsCf(pMap) -> pd.DataFrame:
    ''' read a map of pools in dataframes to a single dataframe, with key as 1st level index'''

    if isinstance(pMap, pd.DataFrame):
        return pMap

    pNames = pMap & lens.Keys().collect()
    pFlows = pMap & lens.Values().collect()
    pColumns = pMap & lens.Values().F(lambda x:x.columns.to_list()).collect()
    headers = tz.concat([  [ (k,c) for c in cs]  for k,cs in zip(pNames,pColumns)])
    headerIndex = pd.MultiIndex.from_tuples(headers)
    df = pd.concat(pFlows,axis=1)
    df.columns = headerIndex
    return df

def readTriggers(tMap) -> pd.DataFrame:
    ''' read a map of triggers in dataframes to a single dataframe, with key as 1st level index'''
    if tMap == {} or tMap is None:
        return pd.DataFrame()
    t = tz.pipe({k:v for k,v in tMap.items() if v}
                ,lambda x: tz.valmap(buildJointCf, x)
                ,lambda x: tz.valfilter( lambda y: not y.empty ,x))
    tKeys = t.keys()
    tVals = t.values()
    return pd.concat(list(tVals), axis=1, keys=tuple(tKeys))


def readInspect(r:dict) -> pd.DataFrame:
    """ read inspect result from waterfall and run assumption input , return a joined dataframe ordered by date"""
    if 'inspect' in r:
        i = r['inspect']
        u = unifyTs(i.values())
    else:
        u = pd.DataFrame()

    if not 'waterfallInspect' in r:
        return pd.concat([u],axis=1).sort_index()
    
    if r['waterfallInspect'] is None:
        return pd.concat([u],axis=1).sort_index()
    
    if r['waterfallInspect'].empty:
        return pd.concat([u],axis=1).sort_index()

    w = r['waterfallInspect'].to_records()
    wdf = tz.pipe(tz.groupby(lambda x:tz.nth(2,x), w)
            ,lambda m: tz.valmap(lambda xs: [ (_[1],_[4]) for _ in xs],m)
            ,lambda m: tz.valmap(lambda xs: pd.DataFrame(xs, columns =['Date', 'val']).set_index('Date'),m)
            ,lambda m: {k:v.rename({"val":k},axis=1) for k,v in m.items() }
            )
    return pd.concat([u,*wdf.values()],axis=1).sort_index()


def readFlowsByScenarios(rs:dict, path, fullName=True) -> pd.DataFrame:
    "read time-series balance flow from multi scenario or mult-structs"
    flows = tz.valmap(lambda x: x & path.get(), rs)
    if fullName:
        flows = tz.itemmap(lambda kv: (kv[0],kv[1].rename(f"{kv[0]}:{kv[1].name}"))   ,flows)
    return pd.concat(flows.values(),axis=1)


def readMultiFlowsByScenarios(rs:dict, _path, fullName=True) -> pd.DataFrame:
    "read multi time-series from multi scenario or mult-structs"
    (path,cols) = _path
    vCols = vList(cols, vStr)
    _flows = tz.valmap(lambda x: x & path.get(), rs)
    flows = tz.valmap(lambda df: df[vCols],_flows)
    scenarioNames = list(flows.keys())
    header = pd.MultiIndex.from_product([ scenarioNames ,vCols], names =['Scenario',"Field"])

    df = pd.concat(list(flows.values()), axis=1)
    df.columns = header
    return df


def readFieldsByScenarios(rs:dict, path, extractor, flip=False) -> pd.DataFrame:
    """  
        read fields from multi scenario or mult-structs
        make sure the `path` points to a single value
    """

    # transfrom result map to values of paths
    tbls = tz.valmap(lambda x: x & path.get(), rs)
    if flip:
        r = tz.valmap(lambda x: x.T.loc.__getitem__(extractor), tbls)
    else:
        r = tz.valmap(lambda x: x.loc.__getitem__(extractor), tbls)
    return pd.DataFrame.from_dict(r)
